/*
 * File:     wx_dialg.cc
 * Purpose:  wxDialogBox and miscellaneous dialog functions
 *
 *                       wxWindows 1.40
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 18-4-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef wx_msw
#include "wx.h"
#pragma hdrstop
#else

#include "common.h"
#include "wx_dialg.h"
#include "wx_utils.h"
#include "wx_frame.h"
#include "wx_panel.h"
#include "wx_item.h"
#include "wx_buttn.h"
#include "wx_messg.h"
#include "wx_choic.h"
#include "wx_check.h"
#include "wx_menu.h"
#include "wx_txt.h"
#include "wx_mtxt.h"
#include "wx_slidr.h"
#include "wx_lbox.h"
#endif


// Dialog box - like panel but doesn't need a frame, and is modal or
// non-modal

wxbDialogBox::wxbDialogBox(void)
{
  __type = wxTYPE_DIALOG_BOX;
}

wxbDialogBox::wxbDialogBox(wxFrame *Parent, char *Title, Bool Modal, 
     int x, int y, int width, int height, int style, Bool create) :
			  wxPanel()
{
  __type = wxTYPE_DIALOG_BOX;
}

void wxbDialogBox::Centre(int direction)
{
  int display_width, display_height, width, height, x, y;
  wxDisplaySize(&display_width, &display_height);

  GetSize(&width, &height);
  GetPosition(&x, &y);

  if (direction == wxBOTH || direction == wxHORIZONTAL)
    x = (int)((display_width - width)/2);
  if (direction == wxBOTH || direction == wxVERTICAL)
    y = (int)((display_height - height)/2);

  SetSize(x, y, width, height);
}


/*
 * Common dialogs code
 *
 */

/*
 * Message centring code
 *
 */

void wxSplitMessage(char *message, wxList *messageList, wxPanel *panel)
{
  char *copyMessage = copystring(message);
  int i = 0;
  int len = strlen(copyMessage);
  char *currentMessage = copyMessage;
  while (i < len)
  {
    while ((i < len) && (copyMessage[i] != '\n')) i ++;
    if (i < len) copyMessage[i] = 0;
    wxMessage *mess = new wxMessage(panel, currentMessage);
    messageList->Append(mess);
    panel->NewLine();

    currentMessage = copyMessage + i + 1;
  }
  delete copyMessage;
}

void wxCentreMessage(wxList *messageList)
{
  // Do the message centering
  wxNode *node = messageList->First();
  while (node)
  {
    wxMessage *mess = (wxMessage *)node->Data();
    mess->Centre();
    node = node->Next();
  }
}
int wxDialogButtonPressed = 0;

void wxDialogOkButton(wxButton& but, wxEvent& event)
{
  wxDialogBox *dialog = (wxDialogBox *)but.GetParent();
  wxDialogButtonPressed = wxOK;
  dialog->Show(FALSE);
}

void wxDialogCancelButton(wxButton& but, wxEvent& event)
{
  wxDialogBox *dialog = (wxDialogBox *)but.GetParent();
  wxDialogButtonPressed = wxCANCEL;
  dialog->Show(FALSE);
}

void wxDialogYesButton(wxButton& but, wxEvent& event)
{
  wxDialogBox *dialog = (wxDialogBox *)but.GetParent();
  wxDialogButtonPressed = wxYES;
  dialog->Show(FALSE);
}

void wxDialogNoButton(wxButton& but, wxEvent& event)
{
  wxDialogBox *dialog = (wxDialogBox *)but.GetParent();
  wxDialogButtonPressed = wxNO;
  dialog->Show(FALSE);
}

// Return NULL if cancel pressed
char *wxGetTextFromUser(char *message, char *caption, char *default_value,
                        wxFrame *parent, int x, int y, Bool centre)
{
  if (x < 0) x = wxDIALOG_DEFAULT_X;
  if (y < 0) y = wxDIALOG_DEFAULT_Y;

  wxDialogBox dialog(parent, caption, TRUE, x, y, wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);
  dialog.SetClientSize(wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);

  wxList messageList;
  wxSplitMessage(message, &messageList, &dialog);

  dialog.NewLine();

  wxText text(&dialog, NULL, NULL, default_value, -1, -1, 15);
  dialog.NewLine();

  wxButton *ok = new wxButton(&dialog, (wxFunction)&wxDialogOkButton, "OK");
  (void)new wxButton(&dialog, (wxFunction)&wxDialogCancelButton, "Cancel");

  wxDialogButtonPressed = 0;

  ok->SetDefault();
  text.SetFocus();

  dialog.Fit();

  if (centre)
    wxCentreMessage(&messageList);

  dialog.Centre();
  dialog.Show(TRUE);

  if (wxDialogButtonPressed == wxOK)
    return text.GetValue();
  else
    return NULL;
}


char *wxGetSingleChoice(char *message, char *caption, int n, char *choices[],
                        wxFrame *parent, int x, int y, Bool centre, int width, int height)
{
  if (x < 0) x = wxDIALOG_DEFAULT_X;
  if (y < 0) y = wxDIALOG_DEFAULT_Y;

  // Really want to centre on screen
  wxDialogBox dialog(parent, caption, TRUE, x, y, wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);
  dialog.SetClientSize(wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);

  wxList messageList;
  wxSplitMessage(message, &messageList, &dialog);

  dialog.NewLine();

  wxListBox listbox(&dialog, NULL, NULL, wxSINGLE,
                     -1, -1, width, height, n, choices);

  dialog.NewLine();
  wxButton *ok = new wxButton(&dialog, (wxFunction)&wxDialogOkButton, "OK");
  ok->SetDefault();

  (void)new wxButton(&dialog, (wxFunction)&wxDialogCancelButton, "Cancel");

  wxDialogButtonPressed = 0;
  dialog.Fit();

  if (centre)
    wxCentreMessage(&messageList);

  dialog.Centre();
  dialog.Show(TRUE);
  if (wxDialogButtonPressed == wxOK)
  {
    int sel = listbox.GetSelection();
    if (sel >= 0)
      return listbox.GetString(sel);
    else return NULL;
  }
  else
    return NULL;
}

int wxGetSingleChoiceIndex(char *message, char *caption, int n, char *choices[],
                           wxFrame *parent, int x, int y, Bool centre,
                           int width, int height)
{
  // Really want to centre on screen
  if (x < 0) x = wxDIALOG_DEFAULT_X;
  if (y < 0) y = wxDIALOG_DEFAULT_Y;


  wxDialogBox dialog(parent, caption, TRUE, x, y, wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);
  dialog.SetClientSize(wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);

  wxList messageList;
  wxSplitMessage(message, &messageList, &dialog);

  dialog.NewLine();

  wxListBox listbox(&dialog, NULL, NULL, wxSINGLE,
                     -1, -1, width, height, n, choices);
  dialog.NewLine();

  wxButton *ok = new wxButton(&dialog, (wxFunction)&wxDialogOkButton, "OK");
  ok->SetDefault();

  (void)new wxButton(&dialog, (wxFunction)&wxDialogCancelButton, "Cancel");

  wxDialogButtonPressed = 0;
  dialog.Fit();

  if (centre)
    wxCentreMessage(&messageList);

  listbox.SetFocus();
  dialog.Centre();
  dialog.Show(TRUE);
  if (wxDialogButtonPressed == wxOK)
  {
    return listbox.GetSelection();
  }
  else
    return -1;
}

char *wxGetSingleChoiceData(char *message, char *caption, int n,
                            char *choices[], char *client_data[],
                            wxFrame *parent, int x, int y, Bool centre,
                            int width, int height)
{
  // Really want to centre on screen
  if (x < 0) x = wxDIALOG_DEFAULT_X;
  if (y < 0) y = wxDIALOG_DEFAULT_Y;

  wxDialogBox dialog(parent, caption, TRUE, x, y, wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);
  dialog.SetClientSize(wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);

  wxList messageList;
  wxSplitMessage(message, &messageList, &dialog);

  dialog.NewLine();

  wxListBox listbox(&dialog, NULL, NULL, wxSINGLE, -1, -1,
                     width, height);
  int i;
  for (i = 0; i < n; i++)
    listbox.Append(choices[i], client_data[i]);

  dialog.NewLine();

  wxButton *ok = new wxButton(&dialog, (wxFunction)&wxDialogOkButton, "OK");
  ok->SetDefault();

  (void)new wxButton(&dialog, (wxFunction)&wxDialogCancelButton, "Cancel");

  wxDialogButtonPressed = 0;
  dialog.Fit();

  if (centre)
    wxCentreMessage(&messageList);

  dialog.Centre();
  dialog.Show(TRUE);
  if (wxDialogButtonPressed == wxOK)
  {
    int sel = listbox.GetSelection();
    if (sel >= 0)
      return listbox.GetClientData(sel);
    else return NULL;
  }
  else
    return NULL;
}

// Pop up a message box
int wxMessageBox(char *message, char *caption, int type,
                 wxFrame *parent, int x, int y, Bool centre)
{
  wxDialogBox dialog(parent, caption, TRUE, x, y, wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);

  wxList messageList;
  wxSplitMessage(message, &messageList, &dialog);

  dialog.NewLine();

  wxButton *ok = NULL;
  wxButton *cancel = NULL;
  wxButton *yes = NULL;
  wxButton *no = NULL;

  if (type & wxYES_NO)
  {
    yes = new wxButton(&dialog, (wxFunction)&wxDialogYesButton, "Yes");
    no = new wxButton(&dialog, (wxFunction)&wxDialogNoButton, "No");
  }

  if (type & wxOK)
  {
    ok = new wxButton(&dialog, (wxFunction)&wxDialogOkButton, "OK");
  }

  if (type & wxCANCEL)
  {
    cancel = new wxButton(&dialog, (wxFunction)&wxDialogCancelButton, "Cancel");
  }

  if (ok)
    ok->SetDefault();
  else if (yes)
    yes->SetDefault();

  dialog.Fit();

  // Do the message centering
  if (centre)
    wxCentreMessage(&messageList);

  if (ok && !cancel && !yes && !no)
    ok->Centre();

  if ((x < 0) || (y < 0))
    dialog.Centre();

  wxDialogButtonPressed = 0;
  dialog.Show(TRUE);

  if (ok)
    delete ok;
  if (cancel)
    delete cancel;
  if (yes)
    delete yes;
  if (no)
    delete no;

  return wxDialogButtonPressed;
}

void wxPopUpSelected(wxListBox& listbox, wxCommandEvent& event)
{
  wxDialogBox *dialog = (wxDialogBox *)listbox.GetParent();
  wxDialogButtonPressed = wxOK;
  dialog->Show(FALSE);
}

char *wxGetSinglePopUp(char *message, char *caption, int n, char *choices[],
                        wxFrame *parent, int x, int y, Bool centre, int width, int height)
{
  if (x < 0) x = wxDIALOG_DEFAULT_X;
  if (y < 0) y = wxDIALOG_DEFAULT_Y;

  // Really want to centre on screen
  wxDialogBox dialog(parent, caption, TRUE, x, y, wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);
  dialog.SetClientSize(wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);

  wxList messageList;
  wxSplitMessage(message, &messageList, &dialog);

  dialog.NewLine();

  wxListBox listbox(&dialog, (wxFunction) wxPopUpSelected, NULL, wxSINGLE,
                     -1, -1, width, height, n, choices);

  dialog.NewLine();

  wxButton *cancel = new wxButton(&dialog, (wxFunction)&wxDialogCancelButton, "Cancel");
  cancel->SetDefault();

  wxDialogButtonPressed = 0;
  dialog.Fit();

  if (centre)
    wxCentreMessage(&messageList);

  listbox.SetFocus();
  dialog.Centre();
  dialog.Show(TRUE);
  if (wxDialogButtonPressed == wxOK)
  {
    int sel = listbox.GetSelection();
    if (sel >= 0)
      return listbox.GetString(sel);
    else return NULL;
  }
  else
    return NULL;
}

int wxGetSinglePopUpIndex(char *message, char *caption, int n, char *choices[],
                           wxFrame *parent, int x, int y, Bool centre,
                           int width, int height)
{
  // Really want to centre on screen
  if (x < 0) x = wxDIALOG_DEFAULT_X;
  if (y < 0) y = wxDIALOG_DEFAULT_Y;


  wxDialogBox dialog(parent, caption, TRUE, x, y, wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);
  dialog.SetClientSize(wxDIALOG_DEFAULT_W, wxDIALOG_DEFAULT_H);

  wxList messageList;
  wxSplitMessage(message, &messageList, &dialog);

  dialog.NewLine();

  wxListBox listbox(&dialog, (wxFunction) wxPopUpSelected, NULL, wxSINGLE,
                     -1, -1, width, height, n, choices);

  dialog.NewLine();

  wxButton *cancel = new wxButton(&dialog, (wxFunction)&wxDialogCancelButton, "Cancel");
  cancel->SetDefault();

  wxDialogButtonPressed = 0;
  dialog.Fit();

  if (centre)
    wxCentreMessage(&messageList);

  listbox.SetFocus();
  dialog.Centre();
  dialog.Show(TRUE);
  if (wxDialogButtonPressed == wxOK)
  {
    return listbox.GetSelection();
  }
  else
    return -1;
}
