/*
 * File:     wx_canvs.cc
 * Purpose:  wxCanvas implementation (MSW)
 *
 *                       wxWindows 1.40
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 18-4-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include "common.h"
#include "wx_frame.h"
#include "wx_dc.h"
#include "wx_canvs.h"
#include "wx_stdev.h"
#include "wx_utils.h"

wxCanvas::wxCanvas(void)
{
  is_retained = FALSE;
  window_parent = NULL;
  handle = NULL;
  horiz_units = 1;
  vert_units = 1;
  xscroll_pixels_per_line = 1;
  yscroll_pixels_per_line = 1;
  xscroll_lines = 0;
  yscroll_lines = 0;
  xscroll_lines_per_page = 0;
  yscroll_lines_per_page = 0;
  xscroll_position = 0;
  yscroll_position = 0;
  wx_dc = NULL;
}

wxCanvas::wxCanvas(wxFrame *frame, int x, int y, int width, int height, int style, Bool create):
  wxbCanvas(frame, x, y, width, height, style)
{
  is_retained = FALSE;
  handle = NULL;
  window_parent = NULL;
  horiz_units = 1;
  vert_units = 1;
  xscroll_pixels_per_line = 1;
  yscroll_pixels_per_line = 1;
  xscroll_lines = 0;
  yscroll_lines = 0;
  xscroll_lines_per_page = 0;
  yscroll_lines_per_page = 0;
  xscroll_position = 0;
  yscroll_position = 0;
  wx_dc = NULL;

  if(create)
	Create(frame, x, y, width, height, style);
}

Bool wxCanvas::Create(wxFrame *frame, int x, int y, int width, int height, int style)
{ int cw, ch;

  if(!frame) return FALSE;

  if(x == -1)
	x = 0;
  if(y == -1)
	y = 0;
  frame->GetClientSize(&cw, &ch);
  if(width == -1 || width > cw)
	width = cw;
  if(height == -1 || height > ch)
	height = ch;
  wxRect srect(x, y, width, height);
  ClientToScreen(srect);

  if(style & wxBORDER)
  { handle = new CursWnd(frame->handle, srect, CW_THIN_LINE, this);
    handle->Border(CW_THIN_LINE);
  }
  else handle = new CursWnd(frame->handle, srect, CW_NO_LINE, this);
  if(!handle) return FALSE;

  wx_dc = new wxCanvasDC(this);
  if(!wx_dc) return FALSE;

  wx_dc->cliprect = GetClientRect();

  frame->AddChild(this);
  window_parent = frame;

  return TRUE;
}

wxCanvas::~wxCanvas(void)
{
  if(wx_dc) delete wx_dc;
}

void wxCanvas::SetColourMap(wxColourMap *cmap)
{
}

/*
 * horizontal/vertical: number of pixels per unit (e.g. pixels per text line)
 * x/y_length:        : no. units per scrollbar
 * x/y_page:          : no. units per page scrolled
 */
void wxCanvas::SetScrollbars(int horizontal, int vertical,
                             int x_length, int y_length,
                             int x_page, int y_page)
{
  horiz_units = horizontal;
  vert_units = vertical;
  xscroll_pixels_per_line = horizontal;
  yscroll_pixels_per_line = vertical;
  xscroll_lines = x_length;
  yscroll_lines = y_length;
  xscroll_lines_per_page = x_page;
  yscroll_lines_per_page = y_page;
}

void wxCanvas::GetScrollUnitsPerPage(int *x_page, int *y_page)
{
  *x_page = xscroll_lines_per_page;
  *y_page = yscroll_lines_per_page;
}

/*
 * Scroll to given position (scroll position, not pixel position)
 */
void wxCanvas::Scroll(int x_pos, int y_pos)
{
  int old_x, old_y;
  int width, height;

  ViewStart(&old_x, &old_y);
  if (((x_pos == -1) || (x_pos == old_x)) && ((y_pos == -1) || (y_pos == old_y)))
    return;

  GetClientSize(&width, &height);
  TheScreen->Update(FALSE);

  if (x_pos > -1)
  { handle->Clear();
    wx_dc->cliprect = wxRect(0, 0, width, height);
    xscroll_position = x_pos;
  }

  if (y_pos > -1)
  { int dy = y_pos - yscroll_position;
    if(abs(dy) >= height)
    { handle->Clear();
      wx_dc->cliprect = wxRect(0, 0, width, height);
    }
    else
    { handle->Scroll(dy); // + is up, - is down
      if(dy > 0)
	wx_dc->cliprect = wxRect(0, dy, width, height);
      else wx_dc->cliprect = wxRect(0, 0, width, height + dy);
    }
    yscroll_position = y_pos;
  }

  OnPaint();
  wx_dc->cliprect = wxRect(0, 0, width, height);
  TheScreen->Update(TRUE);
}

void wxCanvas::EnableScrolling(Bool x_scroll, Bool y_scroll)
{
}

void wxCanvas::GetVirtualSize(int *x, int *y)
{
    *x = xscroll_pixels_per_line * xscroll_lines;
    *y = yscroll_pixels_per_line * yscroll_lines;
}

// Where the current view starts from
void wxCanvas::ViewStart(int *x, int *y)
{
  *x = xscroll_position;
  *y = yscroll_position;
}
