/*
 * File:     wx_dialg.cc
 * Purpose:  wxDialogBox and miscellaneous dialog functions
 *
 *                       wxWindows 1.40
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 18-4-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */



#include "common.h"
#include "wx_dialg.h"
#include "wx_utils.h"
#include "wx_frame.h"
#include "wx_main.h"
#include "wx_lbox.h"
#include "wx_choic.h"
#include "wx_list.h"
#include "wx_txt.h"
#include "wx_buttn.h"
#include "wx_messg.h"
#include "winmgr.h"

extern "C"
{
#include <dirent.h>
}
#ifdef __BORLANDC__
#include <dir.h>
#include <direct.h>
#include <dos.h>
#endif

char *wxXFileSelector(wxFrame *parent, char *path, char *file, char *message, int flags, char *wild_card);

// Dialog box - like panel but doesn't need a frame, and is modal or
// non-modal

wxDialogBox::wxDialogBox(wxFrame *Parent, char *Title, Bool Modal,
		int x, int y, int width, int height, int style, Bool create):
  wxbDialogBox(Parent, Title, Modal, x, y, width, height, style, FALSE)
{
  diag_frame = NULL;
  diag_panel = NULL;
  modal = Modal;
  if(create)
	Create(Parent, Title, Modal, x, y, width, height, style);
}

Bool wxDialogBox::Create(wxFrame *Parent, char *Title, Bool Modal,
		int x, int y, int width, int height, int style)
{
  if(x == -1)
	x = wxDIALOG_DEFAULT_X;
  if(y == -1)
	y = wxDIALOG_DEFAULT_Y;
  if(width == -1)
	width = wxDIALOG_DEFAULT_W;
  if(height == -1)
	height = wxDIALOG_DEFAULT_H;

  diag_frame = new wxFrame(NULL, Title, x, y, width, height);
  if(!diag_frame) return FALSE;

  diag_frame->GetClientSize(&width, &height);
  window_parent = diag_frame;
  diag_panel = new wxPanel(diag_frame, 0, 0, width, height, style);
  if(!diag_panel) return FALSE;
  handle = diag_panel->handle;

  return TRUE;
}

wxDialogBox::~wxDialogBox()
{
  handle = NULL;
  window_parent = NULL;
  if(diag_frame) delete diag_frame;
}

void wxDialogBox::Fit(void)
{ int w, h;

  if(!diag_frame) return;

  wxPanel::Fit();
  wxPanel::GetSize(&w, &h);
  diag_frame->SetClientSize(w,h);
}

void wxDialogBox::SetSize(int x, int y, int width, int height)
{
  if(!diag_frame) return;

  diag_frame->SetSize(x,y,width, height);
  diag_frame->GetClientSize(&width, &height);
  wxPanel::SetSize(0,0,width, height);
}

void wxDialogBox::GetSize(int *width, int *height)
{
  if(!diag_frame) return;

  diag_frame->GetSize(width, height);
}

void wxDialogBox::SetClientSize(int width, int height)
{
  if(!diag_frame) return;

  wxPanel::SetSize(0,0,width, height);
  diag_frame->SetClientSize(width, height);
}

void wxDialogBox::GetPosition(int *x, int *y)
{
  if(!diag_frame) return;

  diag_frame->GetPosition(x,y);
}

void wxDialogBox::Show(Bool show)
{
  if(!diag_frame) return;

  wxPanel::Show(show);
  diag_frame->Show(show);

  if(show)
  { if(modal)
    {
      TheWinMgr->GrabFocus(this);
      wxTheApp->MainLoop();
      TheWinMgr->ReleaseFocus();
    }
  }
  else
  if(modal)
	QuitMainLoop = TRUE; // will be reset by mainloop itself
}

void wxDialogBox::SetTitle(char *title)
{
  if(!diag_frame) return;

  diag_frame->SetTitle(title);
}

char *wxDialogBox::GetTitle(void)
{
  if(!diag_frame) return "";

  return diag_frame->GetTitle();
}

/*
 * Common dialogs code
 *
 */

char fullpathbuf[MAXPATHLEN];

char *wxFileSelector(char *message,
                     char *default_path, char *default_filename, 
                     char *default_extension, char *wildcard, int flags,
                     wxFrame *parent, int x, int y)
{
  char buf[MAXDIRLEN];
  if (!default_path)
    default_path = getcwd(buf, MAXDIRLEN);
  if (!default_filename)
    default_filename = "";

  return wxXFileSelector(parent, default_path, default_filename, message, flags, wildcard);
}

void wxReadDir(wxListBox *filebox, wxListBox *dirbox, char *path, char *filename, DIR *dirp)
{
  filebox->Show(FALSE);
  filebox->Clear();
  dirbox->Show(FALSE);
  dirbox->Clear();
  struct dirent *dir;

  dirbox->Append("..");

  DIR *dirstream = NULL;
  if ((dir=readdir(dirp))==NULL)
  {
	wxMessageBox("There has been an error in opening the directory\n");
	return;
  }

  char buf[MAXPATHLEN];

  // Skip the first directory (.)

  wxStringList files;
  wxStringList dirs;

  while((dir=readdir(dirp))!=NULL)
  {
    strcpy(buf, path);
    int len = strlen(buf);

    if (buf[len - 1] != DIRSEPCHAR)
      strcat(buf, DIRSEPSTR);

    strcat(buf, dir->d_name);
    if (dirstream = opendir(buf))
    {
      closedir(dirstream);

      if (dir->d_name[0] != '.')
        dirs.Add(dir->d_name);
    }
    else if (!wxIsWild(filename) || wxMatchWild(filename, dir->d_name))
      files.Add(dir->d_name);

  }
  files.Sort();
  dirs.Sort();

  wxNode *node = files.First();
  while (node)
  {
    char *s = (char *)node->Data();
    filebox->Append(s);
    delete s;
    delete node;
    node = files.First();
  }

  node = dirs.First();
  while (node)
  {
    char *s = (char *)node->Data();
    dirbox->Append(s);
    delete s;
    delete node;
    node = dirs.First();
  }

  filebox->Show(TRUE);
  dirbox->Show(TRUE);
}

class wxXFileSelDialog: public wxDialogBox
{
 public:
  char *wild_card;
  wxText *name_item;
  wxText *path_item;
  wxListBox *list_item;
  wxListBox *file_item;
  wxListBox *dir_item;
#ifdef __BORLANDC__
  wxChoice *drive_item;
#endif
  wxXFileSelDialog(wxFrame *frame, char *title, Bool modal = FALSE,
              int x = -1, int y = -1, int
              width = -1, int height = -1);
};

wxXFileSelDialog::wxXFileSelDialog(wxFrame *frame, char *title, Bool modal,
              int x, int y, int width, int height):
  wxDialogBox(frame, title, modal, x, y, width, height)
{
}

#ifdef __BORLANDC__

int wxDriveToInt(char *drive)
{  // A=1;B=2;etc.
   return drive[0] - 'A' + 1;
}

char wxIntToDrive(int d)
{ // 1=A;2=B;etc.
  return (char) 'A' + d - 1;
}

// 'Stolen' from Borland's turbo vision
Bool validDrive(char drive)
{
   asm {
	MOV DL, drive
	MOV AH, 36h
	SUB DL, 'A' - 1
	INT 21h
	INC AX
	JE __1
	MOV AL, 1
   }
__1:
  return (Bool) _AX;
}

void wxAddDrives(wxChoice* drivelist)
{ char drive;

  drivelist->Append("A:"); // to avoid drive not ready etc.
  drivelist->Append("B:");
  for(drive='C';drive<='Z';drive++)
  { if(validDrive(drive))
    { wxBuffer[0] = drive;
      wxBuffer[1] = ':';
      wxBuffer[2] = '\0';
      drivelist->Append(wxBuffer);
    }
  }
}
#endif // __BORLANDC__

void wxStripOneDir(char *dir)
{
  int len = strlen(dir);
  if (len > 1)
  {
    int i = len - 1;
    char ch = 0;
    while (ch != DIRSEPCHAR && i > 0)
    {
      ch = dir[i];
      i --;
    }
    i ++;
    if (i == 0)
      i ++;
    dir[i] = 0;
  }
}

void wxXFileSelPath(wxText& text, wxEvent& event)
{
}

void wxXFileSelDirList(wxListBox& listbox, wxEvent& event)
{
  char *name = copystring(listbox.GetStringSelection());
  wxXFileSelDialog *dialog = (wxXFileSelDialog *)listbox.GetParent();
#ifdef __BORLANDC__
  char *drive = copystring(dialog->drive_item->GetStringSelection());
#endif
  char *path = copystring(dialog->path_item->GetValue());
  char buf[MAXPATHLEN];

  strcpy(buf, path);
  int len = strlen(buf);

  if (strcmp(name, "..") == 0)
  {
    wxStripOneDir(buf);
  }
  else
  {
    if (buf[len - 1] != DIRSEPCHAR)
      strcat(buf, DIRSEPSTR);
    strcat(buf, name);
  }

#ifdef __BORLANDC__
  fnmerge(fullpathbuf, drive, buf, NULL, NULL);
#else
  strcpy(fullpathbuf, buf);
#endif

  DIR *dirstream = NULL;
  if (dirstream = opendir(fullpathbuf))
  {
    dialog->path_item->SetValue(buf);
    wxReadDir(dialog->file_item, &listbox, fullpathbuf, dialog->name_item->GetValue(), dirstream);
    closedir(dirstream);
  }
  else dialog->name_item->SetValue(name);

#ifdef __BORLANDC__
  delete drive;
#endif
  delete path;
  delete name;
}

void wxXFileSelFileList(wxListBox& listbox, wxEvent& event)
{
  char *name = copystring(listbox.GetStringSelection());
  wxXFileSelDialog *dialog = (wxXFileSelDialog *)listbox.GetParent();

  dialog->name_item->SetValue(name);
  delete name;
}

#ifdef __BORLANDC__
void wxXFileSelDrive(wxChoice& choice, wxEvent& event)
{
  char *drive = copystring(choice.GetStringSelection());
  wxXFileSelDialog *dialog = (wxXFileSelDialog *)choice.GetParent();

  char path[MAXPATHLEN];
  _getdcwd(wxDriveToInt(drive), fullpathbuf, MAXPATHLEN);
  fnsplit(fullpathbuf, NULL, path, NULL, NULL);

  DIR *dirstream = NULL;
  if (dirstream = opendir(fullpathbuf))
  {
    dialog->path_item->SetValue(path);
    wxReadDir(dialog->file_item, dialog->dir_item, fullpathbuf, dialog->name_item->GetValue(), dirstream);
    closedir(dirstream);
  }
  else
  { wxMessageBox("Error reading drive.");
    drive[0] = wxIntToDrive(_getdrive());
    choice.SetStringSelection(drive);
  }

  delete drive;

}
#endif

int wxXFileSelResponse = 0;
char *wxXFileSelAnswer = NULL;

void wxXFileSelOK(wxButton& ok, wxEvent& event)
{
  wxXFileSelDialog *dialog = (wxXFileSelDialog *)ok.GetParent();
  char *nameval = dialog->name_item->GetValue();
  char *pathval = dialog->path_item->GetValue();
#ifdef __BORLANDC__
  char *driveval = dialog->drive_item->GetStringSelection();
#endif

  if (wxIsWild(nameval))
  {
    DIR *dirstream = NULL;
#ifdef __BORLANDC__
    fnmerge(fullpathbuf, driveval, pathval, NULL, NULL);
#else
    strcpy(fullpathbuf, pathval);
#endif
    if (dirstream = opendir(fullpathbuf))
    {
      wxReadDir(dialog->file_item, dialog->dir_item, fullpathbuf, nameval, dirstream);
      closedir(dirstream);
    }
    return;
  }

#ifdef __BORLANDC__
  fnmerge(fullpathbuf, driveval, pathval, nameval, NULL);
#else
  strcpy(fullpathbuf, pathval);
  int len = strlen(fullpathbuf);
  if (fullpathbuf[len - 1] != DIRSEPCHAR)
    strcat(fullpathbuf, DIRSEPSTR);
  strcat(fullpathbuf, nameval);
#endif

  if (wxXFileSelAnswer)
    delete wxXFileSelAnswer;

  wxXFileSelAnswer = copystring(fullpathbuf);

  dialog->Show(FALSE);
  delete dialog;
  delete nameval;
  delete pathval;

  wxXFileSelResponse = 1;
}

void wxXFileSelCancel(wxButton& cancel, wxEvent& event)
{
  wxXFileSelDialog *dialog = (wxXFileSelDialog *)cancel.GetParent();
  wxXFileSelResponse = 0;
  dialog->Show(FALSE);
  delete dialog;
}

char *wxXFileSelector(wxFrame *parent, char *path, char *file, char *message, int flags, char *wild_card)
{
  wxXFileSelResponse = 0;
  char buf[MAXPATHLEN];
  DIR *dirstream;

  if (!path)
     path = getcwd(fullpathbuf, MAXPATHLEN);
  else
    strcpy(fullpathbuf, path);

#ifdef __BORLANDC__
    char drive[MAXDRIVELEN];
    fnsplit(path, drive, buf, NULL, NULL);  // strip off drive letter and collon
    path = buf;
#endif

  if (!file)
    file = "";

  if ((dirstream=opendir(fullpathbuf))!=NULL)
  {
    wxXFileSelDialog *dialog = new wxXFileSelDialog(parent, "File selector", TRUE, 0, 0, 80, 24);

    (void)new wxMessage(dialog, message);
    dialog->NewLine();
    wxText *name_text = new wxText(dialog, (wxFunction)NULL, "Name", file, -1, -1, 13, -1);
    dialog->NewLine();
    wxText *path_text = new wxText(dialog, (wxFunction)wxXFileSelPath, "Path", path, -1, -1, 35, -1);
    dialog->NewLine();
    wxListBox *filebox = new wxListBox(dialog, (wxFunction)wxXFileSelFileList, "Files", wxSINGLE,
				       -1, -1, 15, 10);
    wxListBox *dirbox = new wxListBox(dialog, (wxFunction)wxXFileSelDirList, "Directories", wxSINGLE,
				      -1, -1, 15, 10);
#ifdef __BORLANDC__
    dialog->NewLine();
    dialog->Tab(35);
    wxChoice *drivelist = new wxChoice(dialog, (wxFunction)wxXFileSelDrive, "Drives", -1, -1, -1, -1);
#endif
    dialog->NewLine();

    (void)new wxButton(dialog, (wxFunction)wxXFileSelOK, "OK");
    (void)new wxButton(dialog, (wxFunction)wxXFileSelCancel, "Cancel");

    dialog->path_item = path_text;
    dialog->name_item = name_text;
#ifdef __BORLANDC__
    dialog->drive_item = drivelist;
    wxAddDrives(drivelist);
    drivelist->SetStringSelection(drive);
#endif
    dialog->dir_item = dirbox;
    dialog->file_item = filebox;
    dialog->wild_card = wild_card;

    dialog->Fit();
    dialog->Centre(wxBOTH);

    char *initial_name = file;
    if (((initial_name == NULL) || (strcmp(initial_name, "") == 0)) &&
        wild_card)
    {
      initial_name = wild_card;
      dialog->name_item->SetValue(initial_name);
    }

    wxReadDir(filebox, dirbox, fullpathbuf, initial_name, dirstream);
    dialog->Show(TRUE);

    if (wxXFileSelResponse == 0)
      return NULL;
    else
    {
      if ((flags & wxOVERWRITE_PROMPT) && FileExists(wxXFileSelAnswer))
      {
	char buf[MAXPATHLEN];
        sprintf(buf, "Overwrite existing file?:\n%s", wxXFileSelAnswer);
        int ans = wxMessageBox(buf, "Warning", wxYES_NO);
        if (ans == wxYES)
          return copystring(wxXFileSelAnswer);
        else return NULL;
      } else return copystring(wxXFileSelAnswer);
    }
  }
  else return NULL;
}


