/*
 * File:     wx_frame.cc
 * Purpose:  wxFrame implementation
 *
 *                       wxWindows 1.40
 * Copyright (c) 1993 Department of Computer Science,
 *                   University of Nijmegen
 *
 *                     Author: Harco de Hilster
 *                       Date: 27-7-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF NIJMEGEN BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include "common.h"
#include "wx_frame.h"
#include "wx_main.h"
#include "wx_utils.h"
#include "curswin.h"
#include "winmgr.h"
#include "wx_list.h"
#include "wx_menu.h"
#include "wx_mbut.h"

#include <signal.h>

wxList wxIconizedWindows;

// INTERNAL MEMBER FUNCTIONS
wxRect wxFrame::GetClientRect(void)
{
  if(wx_menu_bar)
  { wxRect cRect(wxWindow::GetClientRect());
	// The first line is reserved for the menubar
	cRect.Y++;
	cRect.H--;
	return cRect;
  }
  else return wxWindow::GetClientRect();
}

wxRect wxFrame::MinimalWinSize(void)
{ 
  wxRect minrect(GetWinRect());
  if(win_title)
	minrect.W = strlen(win_title) + 5;
  if(wx_menu_bar)
	minrect  = Union(minrect, wx_menu_bar->MinimalWinSize());

  minrect.W += 2;
  if(wx_menu_bar)
	minrect.H = 4;
  else minrect.H = 3;
 
  return minrect;
}


void wxFrame::SetCaption(void)
{
  if(!win_title) return;
  wxRect crect(GetWinRect());

  crect.H = 1;
  crect.X++;
  crect.W = min(crect.W-2, strlen(win_title)+5);

  if (caption == NULL)
  { caption = new CursWnd(handle, crect, 0, NULL);
    caption->Show(TRUE);
  }
  else caption->SetWinRect(crect);

  OnPaint();
}

void wxFrame::OnPaint(void)
{
  caption->Clear();
  caption->Printf(0,0, "[%d][%s]", win_id, win_title);
  TheScreen->InvalidateRect(caption->GetWinRect());
}

Bool wxFrame::ProcessEvent(wxEvent *ev)
{
  if(ev->eventType == wxEVENT_TYPE_WM_COMMAND)
  { wxCommandEvent *c_ev = (wxCommandEvent *) ev;
    switch(c_ev->commandInt)
    { case WM_FOCUSNEXT:	FocusNextSubWin();
				return TRUE;
      case WM_FOCUSPREV:	FocusPreviousSubWin();
				return TRUE;
    }
  }
  else
  if (ev->eventType == wxEVENT_TYPE_CHAR)
  { wxKeyEvent *k_ev = (wxKeyEvent *) ev;
    switch(k_ev->keyCode)
    { 
      case CNTRL('Z'):  kill(0, SIGTSTP);
                        return TRUE;
      case CNTRL('C'):  kill(0, SIGINT);
                        return TRUE;
    }
  };

  return wxWindow::ProcessEvent(ev); 
}

int wxNextWinID(void)
{ int maxid = 0;
  wxNode *node;
  wxFrame *frame;

  node = wxTopLevelWindows.First();

  while(node)
  { frame = (wxFrame *) node->Data();
    if(frame->win_id > maxid)
	maxid = frame->win_id;
    node = node->Next();
  }

  return maxid + 1;
}

// WXWIN MEMBER FUNCTIONS
wxFrame::wxFrame(wxFrame *Parent, char *title, int x, int y,
				 int width, int height, int style):
  wxbFrame(Parent, title, x, y, width, height, style)
{
  frame_type = style & (wxSDI | wxMDI_PARENT | wxMDI_CHILD);
  wx_menu_bar = NULL;
  status_line_exists = FALSE;
  status_text = NULL;
  win_title = NULL;
  icon = NULL;
  modal_showing = FALSE;
  window_parent = NULL; // Frames can't have parents 
  wx_iconized = FALSE;
  win_id = 0; 		// init before searching for higher one
  win_id = wxNextWinID();
  caption = NULL;

  Create(Parent, title, x, y, width, height, style);
}

Bool wxFrame::Create(wxFrame *Parent, char *title, int x, int y,
				 int width, int height, int style)
{
  wxRect winRect(x, y, width, height);
  if(title)
	winRect.W = max(width, strlen(title) + 5 + 2);
  ClientToScreen(winRect);
 
  handle = new CursWnd(NULL, winRect, CW_THICK_LINE, this);
  if(!handle) return FALSE;
  handle->Border(CW_THICK_LINE);
  SetTitle(title);
  return TRUE;
}

wxFrame::~wxFrame(void)
{
  if (handle)
    Show(FALSE);

  if (wx_menu_bar) { delete wx_menu_bar; wx_menu_bar = NULL;};
  if (icon) { delete icon; icon = NULL;};
  if (caption) { delete caption; caption = NULL;};
  if (win_title) { delete win_title; win_title = NULL;};
  if (status_text) { delete status_text; status_text = NULL;};

  if (this == wxTheApp->wx_frame)
	wxTheApp->wx_frame = NULL;
}

void wxFrame::SetSize(int x, int y, int width, int height)
{
  wxbFrame::SetSize(x, y, width, height);
  if (caption) SetCaption();
  if(wx_menu_bar)
  { GetClientSize(&width, &height);
    wx_menu_bar->SetSize(0, -1, width, 1);
  }
}

void wxFrame::SetClientSize(int width, int height)
{
  if(wx_menu_bar)
	wxWindow::SetClientSize(width, height + 1);
  else wxWindow::SetClientSize(width, height);
}

void wxFrame::OnSetFocus(void)
{ wxWindow *win;

  if(caption)
  { caption->WinAttribOn(A_BOLD);
    OnPaint();
  }
  win = FocusedSubWin();
  if(win)
    win->SetFocus();
  Refresh();
}

void wxFrame::OnKillFocus(void)
{
  if (caption)
  { caption->WinAttribOff(A_BOLD);
    OnPaint();
    Refresh();
  }
}

void wxFrame::Iconize(Bool iconize)
{ // Iconizing is simulated by hiding the window. Iconized windows are
  // kept in a list and can be shown in a window menu if desired.
  // Other option is to use character icons and put the icon in the
  // statusbar.

  wx_iconized = iconize;
  if (iconize)
    wxIconizedWindows.Append(this);
  else wxIconizedWindows.DeleteObject(this);

  Show(!iconize);
}

// Equivalent to maximize/restore in Windows
void wxFrame::Maximize(Bool maximize)
{
   Iconize(!maximize);
}

Bool wxFrame::Iconized(void)
{
  return wx_iconized;
}

void wxFrame::SetTitle(char *title)
{
  if(!title) return;
  if(win_title) delete win_title;
  win_title = copystring(title);
  SetCaption();
  OnPaint();
}

char *wxFrame::GetTitle(void)
{
  return (char *) win_title;
}

void wxFrame::SetIcon(wxIcon *wx_icon)
{
  icon = wx_icon;
}

void wxFrame::CreateStatusLine(void)
{
  // All windows share the same statusline
  status_line_exists = TRUE;
}

void wxFrame::SetStatusText(char *text)
{
  if(status_line_exists)
  { if(status_text) delete status_text;
    status_text = text ? copystring(text) : NULL ;
  }

  if(wxFocusFrame == this)
  { if(status_text)
	TheScreen->StatusMsg(status_text);
    else TheScreen->StatusMsg("");
//    TheScreen->Refresh();
  }
}

void wxFrame::LoadAccelerators(char *table)
{
}

void wxFrame::Fit(void)
{
  // Work out max. size
  wxNode *node = children->First();
  int max_width = 0;
  int max_height = 0;
  while (node)
  {
    wxWindow *win = (wxWindow *)node->Data();
    int width, height;
    int x, y;
    win->GetSize(&width, &height);
    win->GetPosition(&x, &y);

    if ((x + width) > max_width)
      max_width = x + width;
    if ((y + height) > max_height)
      max_height = y + height;
    node = node->Next();
  }
  SetClientSize(max_width, max_height);
}

void wxFrame::SetMenuBar(wxMenuBar *menu_bar)
{ int width, height;

  wx_menu_bar = menu_bar;
  // Create the menubar window
  GetClientSize(&width, &height);
  wx_menu_bar->Create(this, 0, -1, width, 1);
  TheScreen->Refresh();
}


