/*
 * File:     wx_item.cc
 * Purpose:  Panel items implementation
 *
 *                       wxWindows 1.40
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 18-4-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


/* When implementing a new item, be sure to:
 *
 * - add the item to the parent panel
 * - set window_parent to the parent
 * - NULL any extra child window pointers not created for this item
 *   (e.g. label control that wasn't needed)
 * - delete any extra child windows in the destructor (e.g. label control)
 * - implement GetSize and SetSize
 * - to find panel position if coordinates are (-1, -1), use GetPosition
 * - call AdvanceCursor after creation, for panel layout mechanism.
 *
 */

#include "common.h"

#include "wx_item.h"
#include "wx_lbox.h"
#include "wx_buttn.h"
#include "wx_mbut.h"
#include "wx_choic.h"
#include "wx_check.h"
#include "wx_messg.h"
#include "wx_slidr.h"
#include "wx_mnuit.h"
#include "wx_menu.h"
#include "wx_txt.h"
#include "wx_mtxt.h"
#include "wx_ptxt.h"

#include "wx_utils.h"
#include "wx_main.h"
#include "wx_frame.h"
#include "wx_dialg.h"
#include "wx_stdev.h"
#include "wx_privt.h"

#include <stdio.h>
#include <ctype.h>
#include <math.h>

// Item members
wxItem::wxItem(void)
{
  item_label = NULL;
  labelPosition = wxHORIZONTAL;
}

wxItem::~wxItem(void)
{
  if(item_label)
	delete item_label;
  item_label = NULL;
}

void wxItem::SetLabel(char *label)
{
  if(item_label)
  { delete item_label;
    item_label = NULL;
  }
  if(label)
	item_label = copystring(label);
}

char *wxItem::GetLabel(void)
{
  return item_label;
}

void wxItem::SetDefault(void)
{
}

void wxItem::OnSetFocus(void)
{
  if(handle)
  { handle->WinAttribOn(A_REVERSE);
    OnPaint();
  }
}

void wxItem::OnKillFocus(void)
{
  if(handle)
  { handle->WinAttribOff(A_REVERSE);
    OnPaint();
  }
}

// Buttons

wxButton::wxButton(wxPanel *panel, wxFunction Function, char *label,
		   int x, int y, int width, int height, Bool create):
  wxbButton(panel, Function, label, x, y, width, height)
{
  if(create)
	Create(panel, Function, label, x, y, width, height);
}

Bool wxButton::Create(wxPanel *panel, wxFunction Function, char *label,
		   int x, int y, int width, int height)
{
  if(!panel) return FALSE;

  SetLabel(label);

  if(x==-1 || y == -1)
	panel->GetValidPosition(&x, &y);
  if(width == -1)
	width = strlen(label)+2;
  if(height == -1)
	height = 1;

  wxRect srect(x, y, width, height);
  ClientToScreen(srect);
  handle = new CursWnd(panel->handle, srect, CW_NO_LINE, this);
  if(!handle) return FALSE;

  handle->Show(TRUE);
  OnPaint();

  panel->AddChild(this);
  Callback(Function);
  panel->AdvanceCursor(this);
  return TRUE;
}

wxButton::~wxButton(void)
{
}

void wxButton::OnPaint()
{ int w,h;

  if(!handle) return;

  handle->Clear();
  GetClientSize(&w, &h);
  handle->Printf(0,0, "(%*s)", w-2, item_label);
}

void wxButton::OnSetFocus(void)
{
  if(handle)
  { handle->WinAttribOn(A_REVERSE);
    OnPaint();
  }
}

void wxButton::OnKillFocus(void)
{
  if(handle)
  { handle->WinAttribOff(A_REVERSE);
    OnPaint();
  }
}

void wxButton::SetDefault(void)
{ wxPanel *panel = (wxPanel *) GetParent();
  wxButton *olddef;

  if(panel)
  {	olddef = panel->defaultItem;
	if(olddef)
	{ olddef->handle->WinAttribOff(A_UNDERLINE); // No longer default
	  olddef->OnPaint();
	};
	panel->defaultItem = this;
  }

  handle->WinAttribOn(A_UNDERLINE);
  OnPaint();
}

Bool wxButton::ProcessEvent(wxEvent *ev)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;
  wxFrame *parent_frame = (wxFrame *) GetGrandParent();

  if(!parent_frame) return FALSE;

  if (k_ev->eventType == wxEVENT_TYPE_CHAR)
  { switch(k_ev->keyCode)
    {
      case KEY_ENTER:	{ wxCommandEvent event(wxEVENT_TYPE_BUTTON_COMMAND);
			  event.eventObject = this;
			  Command(event);
			}
			return TRUE;
      default:		return FALSE;
    }
  };

  return wxItem::ProcessEvent(ev);
}

// Menu button

wxMenuButton::wxMenuButton(wxPanel *panel, wxFunction func,
	wxMenuItem *menuItem, int x, int y, int width, int height, Bool create) :
	wxButton(panel, func, NULL, x, y, width, height, FALSE)
{
  if(create)
	Create(panel, func, menuItem, x, y, width, height);
}

Bool wxMenuButton::Create(wxPanel *panel, wxFunction func,
	wxMenuItem *menuItem, int x, int y, int width, int height)
{
  if(!panel) return FALSE;

  SetLabel(menuItem->itemName);

  menu_item = menuItem;
  menu_item->menu_button = this;

  if((x==-1) || (y==-1))
	panel->GetValidPosition(&x, &y);

  if(width == -1)
	width = strlen(menuItem->itemName);
  if(height == -1)
	height = 1;

  wxRect srect(x,y,width, height);
  ClientToScreen(srect);
  handle = new CursWnd(panel->handle, srect, CW_NO_LINE, this);

  if(!handle) return FALSE;

  handle->Show(TRUE);
  OnPaint();

  if(menuItem->subMenu)
  {  int x, y, px, py;
     wxFrame *parent_frame = (wxFrame *) panel->GetParent();
     panel->GetPosition(&px, &py);
     GetPosition(&x,&y);
     menuItem->subMenu->top_level_menu = menu_item->parent_menu;
     if(labelPosition == wxVERTICAL)
	menuItem->subMenu->Create(parent_frame, px+x, py+y+ height,
				 -1, -1, wxVERTICAL);
     else menuItem->subMenu->Create(parent_frame, px+x+width+1, py+y+1,
				 -1, -1, wxVERTICAL);
  }

  panel->AddChild(this);
  Callback(func);
  panel->AdvanceCursor(this);
  return TRUE;
}

void wxMenuButton::OnPaint(void)
{ int w, h;

  if(!handle) return;

  handle->Clear();

  if((menu_item->menuFlags & MNU_ENABLE))
	handle->WinAttribOn(A_BOLD);
  else handle->WinAttribOff(A_BOLD);

  if(menu_item->itemId == -1)
  { GetClientSize(&w, &h);
    handle->Printf(0,0,"%*.*s",w,w,"------------------------------------");
  }
  else
  if(labelPosition == wxVERTICAL)
	handle->Print(0,0,menu_item->itemName);
  else
  { handle->Print(1,0,menu_item->itemName);
    if(menu_item->menuFlags & MNU_CHECKED)
	handle->Print(0,0, 'X');
    if(menu_item->subMenu)
    { int width, height;
      GetClientSize(&width, &height);
      handle->Print(width-1,0, '>');
    }
  }
}

void wxMenuButton::OnSetFocus(void)
{
  SetDefault();
  wxButton::OnSetFocus();
  wxFrame *parent_frame = (wxFrame *) GetGrandParent();
  if(parent_frame)
	parent_frame->OnMenuSelect(menu_item->itemId);

}

void wxMenuButton::OnKillFocus(void)
{
  wxButton::OnKillFocus();
  wxFrame *parent_frame = (wxFrame *) GetGrandParent();
  if(parent_frame)
	parent_frame->SetStatusText("");
}

void wxMenuButton::SetDefault(void)
{ wxPanel *panel = (wxPanel *) GetParent();
  wxButton *olddef;
 if(panel) panel->defaultItem = this;
/*
  if(panel)
  {	olddef = panel->defaultItem;
	if(olddef)
	{ olddef->handle->WinAttribOff(A_REVERSE); // No longer default
	  olddef->OnPaint();
	};
	panel->defaultItem = this;
  }

  handle->WinAttribOn(A_REVERSE);
  OnPaint();
*/
}

Bool wxMenuButton::ProcessEvent(wxEvent *ev)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;
  wxFrame *parent_frame = (wxFrame *) GetGrandParent();

  if(!parent_frame) return FALSE;

  if (k_ev->eventType == wxEVENT_TYPE_CHAR)
  { switch(k_ev->keyCode)
    {
      case KEY_ENTER:	if(menu_item->subMenu)
				menu_item->subMenu->Show(TRUE);
			else
			if(menu_item->menuFlags & MNU_ENABLE)
			{ menu_item->parent_menu->HideAll();
			  parent_frame->Command(menu_item->itemId);
			}
			return TRUE;
      case KEY_ESC:	menu_item->parent_menu->Hide();
			return TRUE;
      default:		return FALSE;
    }
  };

  return wxButton::ProcessEvent(ev);
}

// Menu Items

wxMenuItem::wxMenuItem(int id, char *label, char *help, wxMenu *submenu)
{
  itemId = id;
  itemName = label ? copystring(label) : NULL;
  helpString = help ? copystring(help) : NULL;
  subMenu = submenu;
  menu_button = NULL;
  parent_menu = NULL;
  menuFlags = MNU_ENABLE;

  if(itemId == -1) // Separator
	menuFlags &= ~MNU_ENABLE;
  if(itemName)
	wxStripMenuCodes(itemName, itemName);
}

wxMenuItem::~wxMenuItem(void)
{
  if(itemName) { delete itemName; itemName = NULL;};
  if(helpString) { delete helpString; helpString = NULL;};
  if(subMenu) { delete subMenu; subMenu = NULL;};
  // the button gets deleted by the window system
}

Bool wxMenuItem::Create(wxPanel *parent, int x, int y, int width,
		 int height, int labelPos)
{
  menu_button = new wxMenuButton(parent, NULL, this, x, y, width, height, FALSE);
  menu_button->SetLabelPosition(labelPos);
  menu_button->Create(parent, NULL, this, x, y, width, height);

  if (!menu_button) return FALSE;
  return TRUE;
}

void wxMenuItem::Enable(Bool flag)
{
  if(flag)
	menuFlags |= MNU_ENABLE;
  else menuFlags &= ~MNU_ENABLE;

  if(menu_button)
  { menu_button->OnPaint();
    menu_button->Refresh();
  }
}

void wxMenuItem::Check(Bool flag)
{
  if(flag)
	menuFlags |= MNU_CHECKED;
  else menuFlags &= ~MNU_CHECKED;

  if(menu_button)
  { menu_button->OnPaint();
    menu_button->Refresh();
  }
}

int wxMenuItem::MaxSize(void)
{  int width;

   if(!itemName)
   {  return 1;
   }

   width = strlen(itemName) ; // space for check mark
   return width;
}

// Menus

// Construct a menu with optional title (then use append)

wxMenu::wxMenu(char *Title, wxFunction func):wxbMenu(Title, func)
{
  no_items = 0;
  menu_bar = NULL;
  title = NULL;
  menu_panel = NULL;
  top_level_menu = this;
}

Bool wxMenu::Create(wxFrame *parent, int x, int y, int width,
		 int height, int labelPos)
{ int i;
  wxNode *node = menuItems.First();
  wxMenuItem *item, *first_item = NULL;

  menu_bar = parent->wx_menu_bar;

  if(labelPos == wxVERTICAL)
  { if(width == -1)
	width = FindMaxSize();
    if(height == -1)
	height = no_items;
    menu_panel = new wxMenuPanel(parent, x, y, width, height, wxBORDER, TRUE);
  }
  else
  { int pw, ph;
    parent->GetClientSize(&pw, &ph);
    if(width == -1)
	width = pw;
    if(height == -1)
	height = 1;
    menu_panel = new wxMenuPanel(parent, x, y, width, height, 0, TRUE);
  }

  if(!menu_panel) return FALSE;

  menu_panel->SetLabelPosition(labelPos);

  // Orientation of menu items is reverse of the panel
  if(labelPos == wxHORIZONTAL)
  { width = -1;
    labelPos = wxVERTICAL;
  }
  else labelPos = wxHORIZONTAL;

  for (i=0;node;node = node->Next(), i++)
  {  item = (wxMenuItem *) node->Data();
     item->parent_menu = this;
     item->menuBar = menu_bar;
     item->Create(menu_panel, -1, -1, width, 1, labelPos);
     if(labelPos == wxHORIZONTAL)
	menu_panel->NewLine();
     if(i==0)
	first_item = item;
  };

  if(first_item)
	first_item->menu_button->SetDefault();
  return TRUE;
}

// The wxWindow destructor will take care of deleting the submenus.
wxMenu::~wxMenu(void)
{ wxNode *node = menuItems.First();
  wxMenuItem *mitem;

  for(;node;node = node->Next())
  { mitem = (wxMenuItem *) node->Data();
    delete mitem;
  }
  // the menu panel gets deleted by the window system
}

// Ordinary menu item
void wxMenu::Append(int Id, char *Label, char *helpString)
{
  wxMenuItem *item = new wxMenuItem(Id, Label, helpString, NULL);
  menuItems.Append(item);
  no_items ++;
}

void wxMenu::AppendSeparator(void)
{

  wxMenuItem *item = new wxMenuItem(-1, NULL, NULL, NULL);
  menuItems.Append(item);
  no_items ++;
}

// Pullright item
void wxMenu::Append(int Id, char *Label, wxMenu *SubMenu, char *helpString)
{
  SubMenu->top_level_menu = top_level_menu;
  wxMenuItem *item = new wxMenuItem(Id, Label, helpString, SubMenu);
  menuItems.Append(item);

  no_items ++;
}

void wxMenu::Enable(int Id, Bool Flag)
{ wxMenuItem *mitem = FindItemForId(Id);

  if(mitem)
	mitem->Enable(Flag);
}

void wxMenu::Check(int Id, Bool Flag)
{ wxMenuItem *mitem = FindItemForId(Id);

  if(mitem)
	mitem->Check(Flag);
}

void wxMenu::Show(Bool flag)
{
  if(menu_panel)
	menu_panel->Show(flag);
}

void wxMenu::SetFocus(void)
{
  if(menu_panel)
	menu_panel->SetFocus();
}

void wxMenu::SetSize(int x, int y, int width, int height)
{
  if(menu_panel)
	menu_panel->SetSize(x, y, width, height);
}

wxRect wxMenu::MinimalWinSize(void)
{ wxRect minrect;

  if(menu_panel)
	minrect = menu_panel->MinimalWinSize();
  return minrect;
}

void wxMenu::Hide(void)
{
  if(menu_panel && (top_level_menu != this))
	menu_panel->Show(FALSE);
}

void wxMenu::HideAll(void)
{
  Hide();
  if(top_level_menu != this)
	top_level_menu->HideAll();
}

int wxMenu::FindMaxSize(void)
{ int maxwidth = 0, width;
  wxNode *node = menuItems.First();
  wxMenuItem *item;

  for(;node;node=node->Next())
  { item = (wxMenuItem *) node->Data();
    if(item)
    { width = item->MaxSize();
      maxwidth = max(maxwidth, width);
    }
  }
  return maxwidth + 2; //checkmark + submenu arrow
}

// Menu Bar
wxMenuBar::wxMenuBar(void)
{
}

wxMenuBar::wxMenuBar(int N, wxMenu *Menus[], char *Titles[]):
		wxbMenuBar(N, Menus, Titles)
{
}

Bool wxMenuBar::Create(wxFrame *parent, int x, int y, int width, int height)
{ int i;

  menu_bar_frame = parent;

  // Create a menu
  menu_bar_menu = new wxMenu();

  // Add the submenus to the menu
  for (i=0;i< n;i++)
  {
    menu_bar_menu->Append(-100 + i, titles[i], menus[i], NULL);
  };

  // Create the menubar window
  menu_bar_menu->Create(parent, x, y, width, height, wxHORIZONTAL);

  if(!menu_bar_menu)
	return FALSE;

  menu_bar_menu->Show(TRUE);
  return TRUE;
}

wxMenuBar::~wxMenuBar(void)
{
  int i;
  for (i = 0; i < n; i++)
  {
	delete titles[i];
  }
  delete menus;
  delete titles;
  if (menu_bar_menu) { delete menu_bar_menu; menu_bar_menu = NULL;};
}

void wxMenuBar::SetFocus(void)
{
  if(menu_bar_menu)
	menu_bar_menu->SetFocus();
}

void wxMenuBar::SetSize(int x, int y, int width, int height)
{
  if(menu_bar_menu)
	menu_bar_menu->SetSize(x, y, width, height);
}

wxRect wxMenuBar::MinimalWinSize(void)
{ wxRect minrect;
  int i;

  for(i=0;i < n; i++)
  { wxStripMenuCodes(titles[i], wxBuffer);
    minrect.W += strlen(wxBuffer) + 1;
  }
  minrect.H = 1;

  return minrect;
}

void wxMenuBar::Append(wxMenu *menu, char *title)
{
  n ++;
  wxMenu **new_menus = new wxMenu *[n];
  char **new_titles = new char *[n];

  int i;
  for (i = 0; i < n - 1; i++)
	{
	  new_menus[i] = menus[i];
	  menus[i] = NULL;
	  new_titles[i] = titles[i];
	  titles[i] = NULL;
	}
  if (menus)
   {
	 delete menus;
	 delete titles;
   }
  menus = new_menus;
  titles = new_titles;

  menus[n-1] = menu;
  titles[n-1] = copystring(title);

  menu->menu_bar = this;
}

// Must only be used AFTER menu has been attached to frame,
// otherwise use individual menus to enable/disable items
void wxMenuBar::Enable(int Id, Bool Flag)
{ wxMenuItem *mitem;

  mitem = FindItemForId(Id);
  if(mitem)
	mitem->Enable(Flag);
}

// Must only be used AFTER menu has been attached to frame,
// otherwise use individual menus
void wxMenuBar::Check(int Id, Bool Flag)
{ wxMenuItem *mitem;

  mitem = FindItemForId(Id);
  if(mitem)
	mitem->Check(Flag);
}

// Single check box item
wxCheckBox::wxCheckBox(wxPanel *panel, wxFunction func, char *Title,
					   int x, int y, int width, int height):
  wxbCheckBox(panel, func, Title, x, y, width, height)
{
  Create(panel, func, Title, x, y, width, height);
}

Bool wxCheckBox::Create(wxPanel *panel, wxFunction func, char *Title,
					   int x, int y, int width, int height)
{
  if(!panel) return FALSE;

  SetLabel(Title);
  checked = FALSE;

  if(x==-1 || y == -1)
	panel->GetValidPosition(&x, &y);
  if(width == -1)
	width = strlen(Title)+4;
  if(height == -1)
	height = 1;

  wxRect srect(x, y, width, height);
  ClientToScreen(srect);
  handle = new CursWnd(panel->handle, srect, CW_NO_LINE, this);
  if(!handle) return FALSE;

  handle->Show(TRUE);
  OnPaint();

  panel->AddChild(this);
  Callback(func);
  panel->AdvanceCursor(this);
  return TRUE;
}

wxCheckBox::~wxCheckBox(void)
{
}

void wxCheckBox::SetValue(Bool val)
{
  checked = val;
  OnPaint();
}

Bool wxCheckBox::GetValue(void)
{
  return checked;
}

void wxCheckBox::OnPaint()
{
  if(!handle) return;

  handle->Clear();
  if(checked)
	handle->Printf(0,0, "[X] %s", item_label);
  else handle->Printf(0,0, "[ ] %s", item_label);
  Refresh();
}

Bool wxCheckBox::ProcessEvent(wxEvent *ev)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;
  if (k_ev->eventType == wxEVENT_TYPE_CHAR)
  { switch(k_ev->keyCode)
    {
      case KEY_ENTER:
      case KEY_SPACE:	{ wxCommandEvent event(wxEVENT_TYPE_CHECKBOX_COMMAND);
			  event.commandInt = !GetValue();
			  event.eventObject = this;
			  Command(event);
			}
			return TRUE;
      default:		return FALSE;
    }
  };

  return wxItem::ProcessEvent(ev);
}

// Choice item
wxChoice::wxChoice(wxPanel *panel, wxFunction func, char *Title,
	int x, int y, int width, int height, int N, char **Choices, Bool create):
  wxbChoice(panel, func, Title, x, y, width, height, N, Choices)
{
  static_label = NULL;
  choice_lbox = NULL;

  if(create)
  	Create(panel, func, Title, x, y, width, height, N, Choices);
}

Bool wxChoice::Create(wxPanel *panel, wxFunction func, char *Title,
		 int x, int y, int width, int height, int N, char **Choices)
{
  if(!panel) return FALSE;

  no_strings = N;
  SetLabel(Title);

  if(x==-1 || y==-1)
	panel->GetValidPosition(&x, &y);

  wxRect srect(x, y, 0, 1);
  ClientToScreen(srect);

  if(Title)
  { srect.W = strlen(Title);
    static_label = new CursWnd(panel->handle, srect, CW_NO_LINE, this);
    if(!static_label) return FALSE;
    static_label->Show(TRUE);
    if(labelPosition == wxHORIZONTAL)
	srect.X += srect.W + 1;
    else srect.Y++;
  }

  if(width == -1)
	  srect.W = 10; // just an initial value
  srect.H = 1;

  handle = new CursWnd(panel->handle, srect, CW_NO_LINE, this);
  if(!handle) return FALSE;
  handle->Show(TRUE);

  srect.X++;srect.Y++;srect.W +=2;srect.H = -1;
  ScreenToClient(srect);
  panel->Freeze(TRUE);  // To reset the panel cursor
  choice_lbox = new wxListBox(panel, NULL, NULL, wxSINGLE, srect.X, srect.Y, srect.W, srect.H, N, Choices);
  if(!choice_lbox) return FALSE;
  // Undo actions of listbox on panel
  panel->Freeze(FALSE);
  choice_lbox->Show(FALSE);
  choice_lbox->WantFocus(FALSE);
  choice_lbox->WantPostProcessEvent(this);

  SetSize(-1, -1, width, height); // To set the width value and listbox

  Callback(func);
  panel->AddChild(this);
  panel->AdvanceCursor(this);

  return TRUE;
}

wxChoice::~wxChoice(void)
{
  if(static_label)
	delete static_label;
}

void wxChoice::Append(char *Item)
{
  no_strings ++;
  choice_lbox->Append(Item);
}

void wxChoice::Clear(void)
{
  no_strings = 0;
  handle->Clear();
  choice_lbox->Clear();
  OnPaint();
}

int wxChoice::GetSelection(void)
{
  return choice_lbox->GetSelection();
}

void wxChoice::SetSelection(int n)
{
  choice_lbox->SetSelection(n);
  OnPaint();
}

int wxChoice::FindString(char *s)
{
  return choice_lbox->FindString(s);
}

char *wxChoice::GetString(int n)
{
  return choice_lbox->GetString(n);
}

void wxChoice::OnSetFocus(void)
{
  if(static_label)
	static_label->WinAttribOn(A_REVERSE);

  if(handle)
	handle->WinAttribOn(A_REVERSE);

  OnPaint();
}

void wxChoice::OnKillFocus(void)
{
  if(static_label)
	static_label->WinAttribOff(A_REVERSE);

  if(handle)
	handle->WinAttribOff(A_REVERSE);

  OnPaint();
}

void wxChoice::OnPaint(void)
{
  if(static_label)
  { static_label->Clear();
    static_label->Print(0,0, item_label);
  }

  if(handle)
  { handle->Clear();
    int sel = GetSelection();
    int w, h;
    handle->GetSize(&w, &h);
    if(sel>=0)
	handle->Print(0,0, GetString(sel));
    handle->Print(A_NORMAL, w-2, 0, " v");
  }
}

void wxChoice::GetPosition(int *x, int *y)
{ wxRect srect;

  if(static_label)
	srect = static_label->GetWinRect();
  else
  if(handle)
	srect = handle->GetWinRect();

  ScreenToClient(srect);
  *x = srect.X;
  *y = srect.Y;
}

void wxChoice::GetSize(int *width, int *height)
{ wxRect srect;

  if(static_label)
	srect = static_label->GetWinRect();
  if(handle)
	srect = Union(srect, handle->GetWinRect());
  if(choice_lbox)
  { wxRect lbrect(choice_lbox->GetWinRect());
    lbrect.H = 0; // don't count the height
    srect = Union(srect, lbrect);
  }

  ScreenToClient(srect);
  *width = srect.W;
  *height = srect.H;
}

void wxChoice::SetSize(int x, int y, int width, int height)
{ int currentX, currentY;

  if(!handle || !choice_lbox) return;

  GetPosition(&currentX, &currentY);
  if(x==-1)
	x = currentX;
  if(y==-1)
	y = currentY;

  wxRect srect(x, y, 0, 1);
  ClientToScreen(srect);

  if(static_label)
  { srect.W = strlen(item_label);
    static_label->SetWinRect(srect);
    if(labelPosition == wxHORIZONTAL)
	srect.X += srect.W + 1;
    else srect.Y++;
  }

  if(width == -1)
	width = choice_lbox->FindMaxSize() + 2; // for arrow down
  srect.W = width;
  srect.H = 1;
  handle->SetWinRect(srect);

  srect.X++;srect.Y++;srect.H = height;
  ScreenToClient(srect);
  choice_lbox->SetSize(srect.X, srect.Y, srect.W, srect.H);

  OnPaint();
}

void wxChoice::Show(Bool flag)
{
  if(static_label)
	static_label->Show(flag);
  if(handle)
	handle->Show(flag);
  if(choice_lbox)
	choice_lbox->Show(flag);
}

Bool wxChoice::ProcessEvent(wxEvent *ev)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;

  if (k_ev->eventType == wxEVENT_TYPE_CHAR)
  { switch(k_ev->keyCode)
    {
      case KEY_ENTER:
      case KEY_DOWN:	choice_lbox->Show(TRUE);
			choice_lbox->SetFocus();
			Refresh();
			return TRUE;
      default:		return FALSE;
    }
  }

  return wxItem::ProcessEvent(ev);
}

Bool wxChoice::PostProcessEvent(wxEvent *ev, Bool handled)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;

  // Post event handling of choice listbox
  if(ev->eventObject == choice_lbox)
  { if (k_ev->eventType == wxEVENT_TYPE_CHAR)
    { switch(k_ev->keyCode)
      { case KEY_ENTER:
        case KEY_SPACE:	{ wxCommandEvent event(wxEVENT_TYPE_CHOICE_COMMAND);
			  event.commandInt = GetSelection();
		  	  event.eventObject = this;
			  event.commandString = copystring(GetStringSelection());
			  Command(event);
			  delete event.commandString;
			  choice_lbox->Show(FALSE);
			  SetFocus();
			  Refresh();
			}
			return TRUE;
	default:	if(!handled)
			{ choice_lbox->Show(FALSE);
			  SetFocus();
			  Refresh();
			  return FALSE;
			}
      }
    }
  }

  return wxItem::ProcessEvent(ev);
}

// EditChoice item
wxEditChoice::wxEditChoice(wxPanel *panel, wxFunction func, char *Title,
	int x, int y, int width, int height, int N, char **Choices, Bool create):
  wxChoice(panel, func, Title, x, y, width, height, N, Choices, create)
{
}

Bool wxEditChoice::PostProcessEvent(wxEvent *ev, Bool handled)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;

  // Post event handling of choice listbox
  if(ev->eventObject == choice_lbox)
  { if (k_ev->eventType == wxEVENT_TYPE_CHAR)
    { switch(k_ev->keyCode)
      { case KEY_EIC:   // fall through
	case KEY_IC:	{ char *newitem = wxGetTextFromUser("New item:");
			  if(newitem)
			  { choice_lbox->Append(newitem);
			    SetStringSelection(newitem);
			    delete newitem;
			    wxCommandEvent event(wxEVENT_TYPE_CHOICE_COMMAND);
			    event.commandInt = GetSelection();
			    event.eventObject = this;
			    event.commandString = copystring(GetStringSelection());
			    Command(event);
			    delete event.commandString;
			  }
			  choice_lbox->Show(FALSE);
		  	  SetFocus();
			  Refresh();
			  return TRUE;
			}
	default:	return wxChoice::PostProcessEvent(ev, handled);
      }
    }
  }

  return wxChoice::PostProcessEvent(ev, handled);
}

// Listbox item
class wxListBoxData : public wxObject
{
public:
  Bool selected;
  char *itemLabel;
  char *clientData;

  wxListBoxData(char *ItemLabel, char *ClientData = NULL)
  { selected = FALSE; itemLabel = copystring(ItemLabel); clientData = ClientData;}

  ~wxListBoxData(void)
  { if(itemLabel) delete itemLabel;}
};

wxListBox::wxListBox(wxPanel *panel, wxFunction func, char *Title, Bool Multiple,
  int x, int y, int width, int height, int N, char **Choices, Bool create):
  wxbListBox(panel, func, Title, Multiple, x, y, width, height, N, Choices)
{
  listItems = new wxList();
  listItems->DeleteContents(TRUE);
  static_label = NULL;
  selected = -1;
  no_selected = 0;
  view_start = 1;
  view_pos = 1;

  if(create)
	Create(panel, func, Title, Multiple, x, y, width, height, N, Choices);
}

Bool wxListBox::Create(wxPanel *panel, wxFunction func,
					   char *Title, Bool Multiple,
					   int x, int y, int width, int height,
					   int N, char **Choices)
{
  if(!panel) return FALSE;

  SetLabel(Title);

  if(x == -1 || y == -1)
	  panel->GetValidPosition(&x, &y);

  Set(N, Choices);

  if(Title)
  { int static_w = strlen(Title);
    wxRect static_rect(x, y, static_w, 1);
    ClientToScreen(static_rect);
    static_label = new CursWnd(panel->handle, static_rect, CW_NO_LINE, this);
    if (!static_label) return FALSE;
    static_label->Show(TRUE);

    if(labelPosition == wxHORIZONTAL)
	x += static_w + 1;
    else y += 1;
  }

  if(width == -1)
	width = max(FindMaxSize(), 5) + 2;
  if(height == -1)
	height = max(2, min(N, 5)) + 2;
  wxRect srect(x, y, width, height);
  ClientToScreen(srect);
  handle = new CursWnd(panel->handle, srect, CW_THIN_LINE, this);
  if(!handle) return FALSE;
  handle->Border(CW_THIN_LINE);
  handle->Show(TRUE);

  OnPaint();
  panel->AddChild(this);
  Callback(func);
  panel->AdvanceCursor(this);
  return TRUE;
}

wxListBox::~wxListBox(void)
{
  if(static_label)
	delete static_label;
  if(listItems)
	delete listItems;
  if(selections)
	delete selections;
}

int wxListBox::FindMaxSize(void)
{ int i;
  int max_len = 0;

  for(i=0;i<no_items;i++)
	max_len = max(max_len, strlen(GetString(i)));
  if(max_len == 0)
	max_len = 5;
  return max_len;
}

void wxListBox::Delete(int N)
{ wxNode *node = listItems->Nth(N); // 0 -> no_items - 1

  no_items --;
  view_start = 1;
  view_pos = 1;
  Deselect(N);  // In case selected
  if(node)
	delete node;
  OnPaint();
}

void wxListBox::Append(char *Item)
{
  Append(Item, NULL);
}

void wxListBox::Append(char *Item, char *Client_data)
{
  no_items ++;
  listItems->Append(new wxListBoxData(Item, Client_data));
  OnPaint();
  Refresh();
}

void wxListBox::Set(int n, char *choices[])
{
  Clear();
  for (int i = 0; i < n; i++)
    listItems->Append(new wxListBoxData(choices[i]));
  no_items = n;

  OnPaint();
  Refresh();
}

int wxListBox::FindString(char *s)
{ wxNode *node = listItems->First();
  wxListBoxData *item;
  int i = 0;

  while(node)
  { item = (wxListBoxData *) node->Data();
    if(strcmp(s, item->itemLabel) == 0)
	return i;
    i++;
    node = node->Next();
  };

  return -1;
}

void wxListBox::Clear(void)
{
  no_items = 0;
  no_selected = 0;
  selected = -1;
  view_start = 1;
  view_pos = 1;
  delete listItems;
  listItems = new wxList();
  listItems->DeleteContents(TRUE);
  if(handle)
	handle->Clear();
  Refresh();
}

void wxListBox::SetSelection(int N)
{ wxNode *node = listItems->Nth(N);
  wxListBoxData *data;

  if(!node) return;

  data = (wxListBoxData *) node->Data();
  if(!data->selected)
  { data->selected = TRUE;
    if(!multiple)
    { Deselect(selected);
      selected = N;
    }
    else no_selected++;
    OnPaint();
    Refresh();
  }
}

void wxListBox::Deselect(int N)
{ wxNode *node = listItems->Nth(N);
  wxListBoxData *data;

  if(!node) return;

  data = (wxListBoxData *) node->Data();

  if(data->selected)
  { if(multiple)
	no_selected--;
    else selected = -1;
    data->selected = FALSE;
  }

  OnPaint();
  Refresh();
}

char *wxListBox::GetClientData(int N)
{ wxNode *node = listItems->Nth(N);
  wxListBoxData *data;

  if(!node)
	return NULL;

  data = (wxListBoxData *) node->Data();
  return data->clientData;
}

// Return number of selections and an array of selected integers
// Use selections field to store data, which will be cleaned up
// by destructor if necessary.
int wxListBox::GetSelections(int **list_selections)
{ wxNode *node = listItems->First();
  wxListBoxData *item;
  int si = 0, ni = 0;

  if(selections)
  { delete selections;
    selections = NULL;
  }

  if(!multiple)
  { selections = new int[1];
    selections[0] = selected;
    *list_selections = selections;
    return 1;
  }
  else
  { selections = new int[no_selected];
    while(node)
    { item = (wxListBoxData *) node->Data();
      if(item->selected)
      { selections[si] = ni;
	si++;
      }
      ni++;
      node = node->Next();
    }
    return si; // should be equal to no_selected
  }
}

// Get single selection, for single choice list items
int wxListBox::GetSelection(void)
{
  return selected;
}

// Find string for position

char *wxListBox::GetString(int N)
{ wxNode *node = listItems->Nth(N);
  wxListBoxData *data;

  wxBuffer[0] = '\0';

  if(!node) return wxBuffer;

  data = (wxListBoxData *) node->Data();
  strcpy(wxBuffer, data->itemLabel);
  return wxBuffer;
}

void wxListBox::OnPaint(void)
{ int i = 1, view_end, w, h, style;
  wxNode *node;
  wxListBoxData *data;

  if(static_label)
  { static_label->Clear();
    static_label->Print(0,0, item_label);
  };

  if(!handle) return;

  handle->Clear();

  GetClientSize(&w, &h);
  view_end = min(no_items - view_start + 1, h);
  node = listItems->Nth(view_start-1); // List counts from zero

  while(node)
  { style = A_NORMAL;
    data = (wxListBoxData *) node->Data();
    if(data->selected)
	style |= A_BOLD;
    if((wxFocusWnd == this) && (i+view_start-1==view_pos))
	style |= A_REVERSE;
    handle->Print(style, 1, i, data->itemLabel);
    if(++i <= view_end)
	node = node->Next();
    else break;
  }
}

/*
   Overload this method if you want to know if the focus of the selection has
   changed. Handy if you scroll though a listbox, don't make a selection, and
   want to show help on each item.
*/
void wxListBox::OnSelectionFocus(int selfocus)
{
}

void wxListBox::OnSetFocus(void)
{
  if(static_label)
  	static_label->WinAttribOn(A_REVERSE);
  OnPaint();
  if(no_items > 0)
    OnSelectionFocus(view_pos - 1);
}

void wxListBox::OnKillFocus(void)
{ int old_view_pos = view_pos;

  if(static_label)
  	static_label->WinAttribOff(A_REVERSE);
  view_pos = -1;
  OnPaint();
  view_pos = old_view_pos;
}

void wxListBox::GetPosition(int *x, int *y)
{ wxRect srect;

  if(static_label)
	srect = static_label->GetWinRect();
  else
  if(handle)
	srect = handle->GetWinRect();

  ScreenToClient(srect);
  *x = srect.X;
  *y = srect.Y;
}

void wxListBox::GetSize(int *width, int *height)
{ wxRect srect;

  if(static_label)
	srect = static_label->GetWinRect();
  if(handle)
	srect = Union(srect, handle->GetWinRect());
  ScreenToClient(srect);
  *width = srect.W;
  *height = srect.H;
}

void wxListBox::GetClientSize(int *width, int *height)
{ // later moet zijn SetClientSize !
  wxWindow::GetClientSize(width, height);
}

void wxListBox::SetSize(int x, int y, int width, int height)
{ int currentX, currentY;

  if(!handle) return;

  GetPosition(&currentX, &currentY);
  if (x == -1)
	x = currentX;
  if (y == -1)
	y = currentY;

  if(static_label)
  { int static_w = strlen(item_label);
    wxRect static_rect(x, y, static_w, 1);
    ClientToScreen(static_rect);
    static_label->SetWinRect(static_rect);

    if(labelPosition == wxHORIZONTAL)
	x += static_w + 1;
    else y += 1;
  }

  if(width == -1)
	width = max(FindMaxSize(), 5) + 2;
  if(height == -1)
	height = max(2, min(no_items, 5)) + 2;
  wxRect srect(x, y, width, height);
  ClientToScreen(srect);
  handle->SetWinRect(srect);

  OnPaint();
  Refresh();
  OnSize(width, height);
}

void wxListBox::Show(Bool flag)
{
  if(!multiple && (selected == -1))
 	SetSelection(0);

  if(static_label)
	static_label->Show(flag);
  wxWindow::Show(flag);
}

void wxListBox::HandleScrollUp(void)
{
  if(view_pos > 1)
  { view_pos--;
    if(view_pos < view_start)
	view_start--;
    OnPaint();
    // Overload if you want listbox item focus change information
    OnSelectionFocus(view_pos - 1);
    Refresh();
  }
  else
	beep();
}

void wxListBox::HandleScrollDown(void)
{ int w, h, view_end;

  if(view_pos >= no_items)
  { beep();
    return;
  }
  view_pos++;

  GetClientSize(&w, &h);
  view_end = view_start + h - 1;
  if(view_pos > view_end)
	view_start++;

  OnPaint();
  OnSelectionFocus(view_pos - 1);
  Refresh();
}

void wxListBox::HandleSelect(void)
{ wxNode *node = listItems->Nth(view_pos - 1);
  wxListBoxData *data;

  if(!node) return;
  data = (wxListBoxData *) node->Data();
  if(data->selected)
  { if(multiple)
	Deselect(view_pos-1);
  }
  else SetSelection(view_pos-1);
}

Bool wxListBox::ProcessEvent(wxEvent *ev)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;
  if (k_ev->eventType == wxEVENT_TYPE_CHAR)
  { switch(k_ev->keyCode)
    {
      case KEY_UP:	HandleScrollUp();
			return TRUE;
      case KEY_DOWN:	HandleScrollDown();
			return TRUE;
      case KEY_ENTER:
      case KEY_SPACE:	{ HandleSelect();
			  wxCommandEvent event(wxEVENT_TYPE_LISTBOX_COMMAND);
			  if (multiple != wxMULTIPLE)
			  { int sel = GetSelection();
                            if(sel<0)
				return TRUE;
			    event.commandInt = sel;
			    event.clientData = GetClientData(event.commandInt);
			    event.commandString = copystring(GetStringSelection());
			  }
			  event.eventObject = this;
			  ProcessCommand(event);
			  if (multiple != wxMULTIPLE)
				delete event.commandString;
			}
			return TRUE;
	default:	return FALSE;
    }
  };

  return wxItem::ProcessEvent(ev);
}

// Message
wxMessage::wxMessage(wxPanel *panel, char *label, int x, int y, Bool create):
  wxbMessage(panel, label, x, y)
{
  if(create)
	Create(panel, label, x, y);
}

Bool wxMessage::Create(wxPanel *panel, char *label, int x, int y)
{
  if(!panel) return FALSE;

  SetLabel(label);

  if(x == -1 || y == -1)
	panel->GetValidPosition(&x, &y);
  int width = strlen(label);

  wxRect srect(x, y, width, 1);
  ClientToScreen(srect);
  handle = new CursWnd(panel->handle, srect, CW_NO_LINE, this);
  if (!handle) return FALSE;
  handle->Show(TRUE);
  WantFocus(FALSE);
  OnPaint();

  panel->AddChild(this);
  panel->AdvanceCursor(this);
  return TRUE;
}

void wxMessage::OnPaint(void)
{
  if(!handle) return;
  handle->Clear();
  handle->Print(0,0, item_label);
}

wxMessage::~wxMessage(void)
{
}

// Text item
wxText::wxText(void)
{
  static_label = NULL;
  view_x = 0;
  cur_x = 1;
  insert = TRUE; // insert mode
  changed = FALSE;
}

wxText::wxText(wxPanel *panel, wxFunction Function, char *label, char *value,
			   int x, int y, int width, int height, Bool create):
  wxbText(panel, Function, label, value, x, y, width, height)
{
  static_label = NULL;
  view_x = 0;
  cur_x = 1;
  insert = TRUE; // insert mode
  changed = FALSE;

  if(create)
	Create(panel, Function, label, value, x, y, width, height);
}

Bool wxText::Create(wxPanel *panel, wxFunction func, char *label, char *value,
			   int x, int y, int width, int height)
{
  if(!panel) return FALSE;
  SetLabel(label);

  if(x == -1 || y == -1)
	  panel->GetValidPosition(&x, &y);

  if(label)
  { int static_w = strlen(label);
    wxRect static_rect(x, y, static_w, 1);
    ClientToScreen(static_rect);
    static_label = new CursWnd(panel->handle, static_rect, CW_NO_LINE, this);
    if (!static_label) return FALSE;
    static_label->Show(TRUE);

    if(labelPosition == wxHORIZONTAL)
	x += static_w + 1;
    else y += 1;
  }

  if(width == -1)
	width = 12;
  if(height == -1)
	height = 1;
  wxRect srect(x, y, width, height);
  ClientToScreen(srect);
  handle = new CursWnd(panel->handle, srect, CW_NO_LINE, this);
  if(!handle) return FALSE;
  handle->Show(TRUE);
  handle->SetCursor(CUR_NORM);

  SetValue(value);

  Callback(func);
  panel->AddChild(this);
  panel->AdvanceCursor(this);
  return TRUE;
}

wxText::~wxText(void)
{
  if(static_label)
	delete static_label;
}

char *wxText::GetValue(void)
{ char *value = NULL;

  value = copystring(text_str.FromPos(0));
  return value;
}

void wxText::SetValue(char *value)
{
  if(value)
	text_str.Set(value);
  else text_str.Set("");

  view_x = 0;
  cur_x = 1;
  changed = FALSE;
  HandleScrollRight(KEY_END);
  Refresh();
}

void wxText::GetPosition(int *x, int *y)
{ wxRect srect;

  if(!handle) return;

  if(static_label)
	srect = static_label->GetWinRect();
  else srect = handle->GetWinRect();

  ScreenToClient(srect);
  *x = srect.X;
  *y = srect.Y;
}

void wxText::GetSize(int *width, int *height)
{ wxRect srect;

  if(static_label)
	srect = static_label->GetWinRect();
  if(handle)
	srect = Union(srect, handle->GetWinRect());
  ScreenToClient(srect);
  *width = srect.W;
  *height = srect.H;
}

void wxText::OnPaint(void)
{
  if(static_label)
  { static_label->Clear();
    static_label->Print(0,0, item_label);
  }

  if(!handle) return;

  handle->Clear();
  if(view_x > 0)
	handle->Print(0,0, '<');
  else handle->Print(0,0, '[');

  handle->Print(1,0, text_str.FromPos(view_x));

  wxRect srect(handle->GetWinRect());
  if(view_x + srect.W - 2 < text_str.Length())
	handle->Print(srect.W-1,0, '>');
  else handle->Print(srect.W-1,0, ']');

  handle->GotoXY(cur_x, 0);
}

void wxText::OnSetFocus(void)
{
  if(static_label)
  { static_label->WinAttribOn(A_REVERSE);
    OnPaint();
  }
}

void wxText::OnKillFocus(void)
{
  if(static_label)
  { static_label->WinAttribOff(A_REVERSE);
    OnPaint();
  }
}

void wxText::HandleScrollLeft(int how)
{ int str_pos;
  int w, h;
  int len;

  if(how == KEY_HOME)
  { view_x = 0;
    cur_x = 1;
  }
  else // KEY_LEFT
  { if(cur_x<=1)
    { 
      str_pos = view_x + cur_x - 1;
      len = text_str.Length();
      if(str_pos == len && len > 0)
      { handle->GetClientSize(&w, &h);
        view_x = max(0, len - w + 2);
        cur_x = str_pos - view_x;
      }
      else view_x = max(0, view_x - 1);
    }
    else cur_x--;
  }

  OnPaint();
}

void wxText::HandleScrollRight(int how)
{ int len;

  if(!handle)
	return;

  wxRect srect = handle->GetWinRect();
  len = text_str.Length();

  if(how == KEY_END)
  {  view_x = max(0, len - srect.W + 2);
     cur_x = min(max(len+1,1), srect.W - 1);
  }
  else  // KEY_RIGHT
  { if(view_x+cur_x > len)
	return;
    if(cur_x>=srect.W-1)
    { if(view_x + srect.W - 2 < len)  // two needed for 'borders'
	view_x++;
    }
    else cur_x++;
  }

  OnPaint();
}

void wxText::HandleChar(int c)
{ int str_pos = view_x + cur_x - 1;

  if(insert)
	text_str.Insert(str_pos, (char) c);
  else text_str.Put(str_pos, (char) c);
  changed = TRUE;

  HandleScrollRight(KEY_RIGHT);
}

void wxText::HandleDelete(int how)
{ int str_pos;

  str_pos = view_x+cur_x-1;

  if(how == KEY_BACKSPACE)
	HandleScrollLeft(KEY_LEFT);
  else
  if(str_pos == text_str.Length()) 
	/* Handle delete at end of string as a backspace */
	HandleScrollLeft(KEY_LEFT);

  str_pos = view_x+cur_x-1;
  text_str.Delete(str_pos);
  changed = TRUE;

  if(cur_x == 1 && str_pos == text_str.Length())
  { HandleScrollLeft(KEY_LEFT);
    HandleScrollRight(KEY_RIGHT);
  }
  else OnPaint();
}

Bool wxText::ProcessEvent(wxEvent *ev)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;
  if (k_ev->eventType == wxEVENT_TYPE_CHAR)
  { switch(k_ev->keyCode)
    { case KEY_LEFT:    HandleScrollLeft(KEY_LEFT);
			Refresh();
			return TRUE;
      case KEY_RIGHT:   HandleScrollRight(KEY_RIGHT);
			Refresh();
			return TRUE;
      case KEY_HOME:	HandleScrollLeft(KEY_HOME);
			Refresh();
			return TRUE;
      case KEY_END:	HandleScrollRight(KEY_END);
			Refresh();
			return TRUE;
      case KEY_BACKSPACE: HandleDelete(KEY_BACKSPACE);
			Refresh();
			return TRUE;
      case KEY_DEL:	/* delete */
      case KEY_DC:      HandleDelete(KEY_DC);
			Refresh();
			return TRUE;
      case KEY_IC:      // fall through
      case KEY_EIC:	insert = !insert;
			if(insert)
				handle->SetCursor(CUR_NORM);
			else handle->SetCursor(CUR_INSERT);
			Refresh();
			return TRUE;
      case KEY_TAB:
      case KEY_ENTER:   if(changed)
			{ wxCommandEvent event(wxEVENT_TYPE_TEXT_COMMAND);
			  event.commandString = GetValue();
			  event.eventObject = this;
			  ProcessCommand(event);
			  delete event.commandString;
			}
			/* let panel handle tab and enter */
			return FALSE;
      default:		if(isprint(k_ev->keyCode))
			{ HandleChar(k_ev->keyCode);
			  Refresh();
			  return TRUE;
			}
			else return FALSE;
    }
  };

  return wxItem::ProcessEvent(ev);
}

// Password text item
wxPasswdText::wxPasswdText(wxPanel *panel, wxFunction Function, char *label, char *value,
	 char echochar, int x, int y, int width, int height, Bool create)
   : wxText(panel, Function, label, value, x, y, width, height, TRUE)
{
  echoch = echochar;
}

void wxPasswdText::OnPaint(void)
{ int i;

  if(static_label)
  { static_label->Clear();
    static_label->Print(0,0, item_label);
  }

  if(!handle) return;

  handle->Clear();
  if(view_x > 0)
	handle->Print(0,0, '<');
  else handle->Print(0,0, '[');

  int txtlen = strlen(text_str.FromPos(view_x));
  for(i=1;i<=txtlen; i++)
	handle->Print(i,0, echoch);

  wxRect srect(handle->GetWinRect());
  if(view_x + srect.W - 2 < text_str.Length())
	handle->Print(srect.W-1,0, '>');
  else handle->Print(srect.W-1,0, ']');

  handle->GotoXY(cur_x, 0);
}

// Multi-line Text item
wxMultiText::wxMultiText(wxPanel *panel, wxFunction Function, char *label, char *value,
			   int x, int y, int width, int height, Bool create) :
     wxbMultiText(panel, Function, label, value, x, y, width, height, FALSE)
{
  view_x = view_y = 0;
  cur_x = 1;
  cur_y = 1;
  insert = TRUE;
  changed = FALSE;
  editable = TRUE;
  lines = new wxList();
  lines->DeleteContents(TRUE);

  if(create)
	Create(panel, Function, label, value, x, y, width, height);
}

Bool wxMultiText::Create(wxPanel *panel, wxFunction func, char *label, char *value,
			   int x, int y, int width, int height)
{
  if(!panel) return FALSE;

  SetLabel(label);

  if(x == -1 || y == -1)
	  panel->GetValidPosition(&x, &y);

  if(label)
  { int static_w = strlen(label);
    wxRect static_rect(x, y, static_w, 1);
    ClientToScreen(static_rect);
    static_label = new CursWnd(panel->handle, static_rect, CW_NO_LINE, this);
    if (!static_label) return FALSE;
    static_label->Show(TRUE);

    if(labelPosition == wxHORIZONTAL)
	x += static_w + 1;
    else y += 1;
  }

  if(width == -1)
	width = 12;
  if(height == -1)
	height = 4;
  wxRect srect(x, y, width, height);
  ClientToScreen(srect);
  handle = new CursWnd(panel->handle, srect, CW_THIN_LINE, this);
  if(!handle) return FALSE;
  handle->Show(TRUE);
  handle->SetCursor(CUR_NORM);

  SetValue(value);

  Callback(func);
  panel->AddChild(this);
  panel->AdvanceCursor(this);
  return TRUE;
}

wxMultiText::~wxMultiText(void)
{
  if(static_label) delete static_label;
  if(lines) delete lines;
}

int wxMultiText::FindMaxStrSize(void)
{ int len = 0;
  wxNode *node;
  TextString *line;

  if(!lines)
	return 0;

  node = lines->First();
  while(node)
  { line = (TextString *) node->Data();
    len += line->Length() + 1; // one extra for the '\n'
    node = node->Next();
  }

  return len;
}

int wxMultiText::NoLines(void)
{
  if(!lines) return 0;
  else return lines->Number();
}

char *wxMultiText::GetValue(void)
{ char *value = NULL;
  TextString *line;

  wxBuffer[0] = '\0';

  int len = FindMaxStrSize();
  if(len == 0)
    return copystring(wxBuffer);

  value = new char[len+1];
  if(!value)
    return copystring(wxBuffer);

  value[0] = '\0';
  wxNode *node = lines->First();
  while(node)
  { line = (TextString *) node->Data();
    strcat(value, line->FromPos(0));
    if(node->Next())
	strcat(value, "\n");
    node = node->Next();
  }
  value[len] = '\0';
  return value;
}

void wxMultiText::GetValue(char *buffer, int maxSize)
{ char *value = GetValue();

  if(!value)
  { buffer[0] = '\0';
    return;
  }

  strncpy(buffer, value, maxSize);
  delete value;
}

void wxMultiText::SetValue(char *value)
{ char *line, *acopy;

  if(lines) delete lines;
  lines = new wxList();
  if(!lines) return;
  lines->DeleteContents(TRUE);
  lines->Append(new TextString());
  
  if(!value) return;

  acopy = copystring(value);
  line = acopy;
  while(*line)
  { char *end = line;
    while(*end && (*end != '\n')) end++;
    TextString *lline = (TextString *)lines->Last()->Data();
    int llen = lline->Length();
    
    if(*end == '\n')
    { *end = '\0';
      lline->Put(llen, line);
      lines->Append(new TextString());
      line = ++end;
    }
    else
    { *end = '\0';
      lline->Put(llen, line);
      break;
    }
  }
  delete acopy;

  view_x = view_y = 0;
  cur_x = 1;
  cur_y = 1;
  changed = FALSE;
  OnPaint();
  Refresh();
}

void wxMultiText::Editable(Bool yn)
{
  editable = yn;
}

Bool wxMultiText::Editable(void)
{
  return editable;
}

void wxMultiText::DeleteLine(int lineno)
{ wxNode *node;

  if(!lines) return;

  node = lines->Nth(lineno);
  if(!node) return;
  delete node; // deletes the TextString too
}

void wxMultiText::Append(char *value)
{ char *line, *acopy;

  if(!value) return;

  if(!lines)
  { lines = new wxList();
    if(!lines) return;
    lines->DeleteContents(TRUE);
    lines->Append(new TextString());
  }

  acopy = copystring(value);
  line = acopy;
  while(*line)
  { char *end = line;
    while(*end && (*end != '\n')) end++;
    TextString *lline = (TextString *)lines->Last()->Data();
    int llen = lline->Length();
    
    if(*end == '\n')
    { *end = '\0';
      lline->Put(llen, line);
      lines->Append(new TextString());
      line = ++end;
    }
    else
    { *end = '\0';
      lline->Put(llen, line);
      break;
    }
  }

  delete acopy;

  changed = FALSE;
  HandleScrollDown(KEY_NPAGE);
  HandleScrollRight(KEY_END);
  OnPaint();
  Refresh();
}

void wxMultiText::GetPosition(int *x, int *y)
{ wxRect srect;

  if(!handle) return;

  if(static_label)
	srect = static_label->GetWinRect();
  else srect = handle->GetWinRect();

  ScreenToClient(srect);
  *x = srect.X;
  *y = srect.Y;
}

void wxMultiText::GetSize(int *width, int *height)
{ wxRect srect;

  if(static_label)
	srect = static_label->GetWinRect();
  if(handle)
	srect = Union(srect, handle->GetWinRect());
  ScreenToClient(srect);
  *width = srect.W;
  *height = srect.H;
}

void wxMultiText::OnPaint(void)
{ TextString *texts;

  if(!handle) return;

  if(static_label)
  { static_label->Clear();
    static_label->Print(0,0, item_label);
  }

  if(!lines) return;

  handle->Clear();
  wxRect srect(handle->GetClientRect());
  wxNode *node = lines->Nth(view_y);
  int line_no = 1;

  while(node && (line_no <= srect.H))
  { texts = (TextString *) node->Data();
    handle->Print(1,line_no, texts->FromPos(view_x));
    line_no++;
    node = node->Next();
  }

  handle->GotoXY(cur_x, cur_y);
}

void wxMultiText::HandleScrollLeft(int how)
{
  if(how == KEY_HOME)
  { view_x = 0;
    cur_x = 1;
  }
  else // KEY_LEFT
  { if(cur_x<=1)
    { if(view_x > 0)
	view_x--;
      else return;
    }
    else cur_x--;
  }
}

void wxMultiText::HandleScrollRight(int how)
{
  if(!handle || !lines)
	return;

  wxNode *node = lines->Nth(view_y+cur_y-1);
  if(!node)
	return;
  TextString *line = (TextString *) node->Data();
  int len = line->Length();
  wxRect srect = handle->GetClientRect();

  if(how == KEY_END)
  {  view_x = max(0, len - srect.W + 1);
     cur_x = min(len+1, srect.W);
  }
  else  // KEY_RIGHT
  { if(view_x+cur_x > len)
	return;
    if(cur_x>=srect.W)
    { if(view_x + srect.W <= len)
	view_x++;
      else return;
    }
    else cur_x++;
  }
}

void wxMultiText::HandleScrollUp(int how)
{
  wxRect srect(handle->GetClientRect());

  if(how == KEY_PPAGE)
  { view_y = max(0, view_y - srect.H);  // scroll one page back
  }
  else // KEY_UP
  { if(cur_y<=1)
    { if(view_y > 0)
	view_y--;
      else return;
    }
    else cur_y--;

    wxNode *node = lines->Nth(view_y+cur_y-1);
    if(!node) return;

    TextString *texts = (TextString *) node->Data();
    if(view_x + cur_x > texts->Length())
    { HandleScrollRight(KEY_END);
      return;
    }
  }
}

void wxMultiText::HandleScrollDown(int how)
{
  wxRect srect(handle->GetClientRect());
  int no_lines = lines->Number();

  if(how == KEY_NPAGE)
  { if((no_lines - view_y) > srect.H)
	view_y += srect.H - 1;  // scroll one page forward
  }
  else // KEY_DOWN
  { if(cur_y >= srect.H)
    { if(view_y < no_lines-srect.H)
	view_y++;
      else return;
    }
    else cur_y++;

    wxNode *node = lines->Nth(view_y+cur_y-1);
    if(!node) return;

    TextString *texts = (TextString *) node->Data();
    if(view_x + cur_x > texts->Length())
    { HandleScrollRight(KEY_END);
      return;
    }
  }
}

void wxMultiText::HandleChar(int c)
{ int str_pos = view_x + cur_x - 1;

  if(!lines || !editable) return;

  wxNode *node = lines->Nth(view_y + cur_y - 1);
  if(!node) return;
  TextString *line = (TextString *) node->Data();

  if(c == KEY_ENTER)
  { if(!insert)
	HandleScrollDown(KEY_DOWN);
    else  // insert mode
    if(str_pos == 0) // insert new line before current
	lines->Insert(node, new TextString());
    else
    if(str_pos == line->Length()) // insert new line after current
    { if(node->Next())
      { node = node->Next();
	lines->Insert(node, new TextString());
      }
      else lines->Append(new TextString());
    }
    else //split line in two
    { TextString *nts;
      char *nl;
      nts = new TextString(line->SubString(0, str_pos-1));
      nl = line->SubString(str_pos, line->Length()-1);
      delete line;
      node->SetData(nts);
      if(node->Next())
      { node = node->Next();
	lines->Insert(node, new TextString(nl));
      }
      else lines->Append(new TextString(nl));
    }
    HandleScrollDown(KEY_DOWN);
    HandleScrollLeft(KEY_HOME);
  }
  else // other then ENTER
  { if(insert)
	line->Insert(str_pos, (char) c);
    else line->Put(str_pos, (char) c);
    HandleScrollRight(KEY_RIGHT);
  }
  changed = FALSE;
}

void wxMultiText::HandleDelete(int how)
{ int str_pos;

  if(!lines || !editable) return;

  str_pos = view_x+cur_x-1;

  if(how == KEY_BACKSPACE)
  { if(str_pos > 0)
	HandleScrollLeft(KEY_LEFT);
    else
    if(view_y+cur_y-1 > 0)
    { HandleScrollUp(KEY_UP);
      HandleScrollRight(KEY_END);
    }
  }

  str_pos = view_x+cur_x-1;

  wxNode *node = lines->Nth(view_y+cur_y-1);
  if(!node) return;
  TextString *line = (TextString *) node->Data();

  if(line->Length() == 0)
  { if(node->Next())
    { delete node;
      HandleScrollUp(KEY_UP);
    }
  }
  else
  if(line->Length() == str_pos)
  { if(node->Next())
    { TextString *nline = (TextString *) node->Next()->Data();
      line->Put(str_pos, nline->FromPos(0));
      delete node->Next();
    }
  }
  else
	line->Delete(str_pos);
  changed = FALSE;
}

Bool wxMultiText::ProcessEvent(wxEvent *ev)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;
  Bool handled = TRUE;

  if (k_ev->eventType == wxEVENT_TYPE_CHAR)
  { switch(k_ev->keyCode)
    { case KEY_LEFT:    HandleScrollLeft(KEY_LEFT);
			break;
      case KEY_RIGHT:   HandleScrollRight(KEY_RIGHT);
			break;
      case KEY_UP:	HandleScrollUp(KEY_UP);
			break;
      case KEY_PPAGE:	HandleScrollUp(KEY_PPAGE);
			break;
      case KEY_DOWN:	HandleScrollDown(KEY_DOWN);
			break;
      case KEY_NPAGE:	HandleScrollDown(KEY_NPAGE);
			break;
      case KEY_HOME:	HandleScrollLeft(KEY_HOME);
			break;
      case KEY_END:	HandleScrollRight(KEY_END);
			break;
      case KEY_BACKSPACE: HandleDelete(KEY_BACKSPACE);
			break;
      case KEY_DEL:	/* fall through */
      case KEY_DC:      HandleDelete(KEY_DC);
			break;
      case KEY_IC:      // fall through
      case KEY_EIC:	insert = !insert;
			if(insert)
				handle->SetCursor(CUR_NORM);
			else handle->SetCursor(CUR_INSERT);
			break;
#ifdef ALT_ENTER
      case ALT_ENTER:	// fall through
#endif
      case CNTRL('J'):	HandleChar(KEY_ENTER);
			break;
      case KEY_TAB:
      case KEY_ENTER:   if(changed)
			{ wxCommandEvent event(wxEVENT_TYPE_TEXT_COMMAND);
			event.commandString = GetValue();
			event.eventObject = this;
			ProcessCommand(event);
			delete event.commandString;
			}
			// let panel handle tab and enter
			handled = FALSE;
			break;
      default:		if(isprint(k_ev->keyCode))
				HandleChar(k_ev->keyCode);
			else handled = FALSE;
			break;
    }

    if(handled)
    { OnPaint();
      Refresh();
    }
    return handled;
  }

  return wxItem::ProcessEvent(ev); // not a char event 
}

// Slider

wxSlider::wxSlider(wxPanel *panel, wxFunction func, char *label, int value,
		   int min_value, int max_value, int width, int x, int y, Bool create):
  wxbSlider(panel, func, label, value, min_value, max_value, width, x, y)
{
  static_handle = NULL;
  value_handle = NULL;
  item_value = 0;
  s_min = s_max = page_size = 0;

  if(create)
	Create(panel, func, label, value, min_value, max_value, width, x, y);
}

Bool wxSlider::Create(wxPanel *panel, wxFunction func, char *label, int value,
		   int min_value, int max_value, int width, int x, int y)
{ int len;

  if(!panel) return FALSE;

  SetLabel(label);

  s_min = min(min_value, max_value);
  s_max = max(min_value, max_value);

  if(x == -1 || y == -1)
	  panel->GetValidPosition(&x, &y);

  if(label)
  { int static_w = strlen(label);
    wxRect static_rect(x, y, static_w, 1);
    ClientToScreen(static_rect);
    static_handle = new CursWnd(panel->handle, static_rect, CW_NO_LINE, this);
    if (!static_handle) return FALSE;
    static_handle->Show(TRUE);

    if(labelPosition == wxHORIZONTAL)
	x += static_w + 1;
    else y += 1;
  }

  char *tmp = IntToString(max_value);
  len = strlen(tmp);
  delete tmp;
  tmp = IntToString(min_value);
  len = max(len, strlen(tmp));
  delete tmp;
  wxRect vrect(x, y, len, 1);
  ClientToScreen(vrect);
  value_handle = new CursWnd(panel->handle, vrect, CW_NO_LINE, this);
  if(!value_handle) return FALSE;
  value_handle->Show(TRUE);

  x += vrect.W+1;
  if(width == -1)
	width = 10;

  wxRect srect(x, y, width+2, 1);
  ClientToScreen(srect);
  handle = new CursWnd(panel->handle, srect, CW_NO_LINE, this);
  if(!handle) return FALSE;
  handle->Show(TRUE);

  SetValue(value);

  Callback(func);
  panel->AddChild(this);
  panel->AdvanceCursor(this);
  return TRUE;
}


wxSlider::~wxSlider(void)
{
  if(static_handle)
	delete static_handle;
  if(value_handle)
	delete value_handle;
}

int wxSlider::GetValue(void)
{
  return item_value;
}

void wxSlider::SetValue(int value)
{
  if(value < s_min)
	item_value = s_min;
  else
  if(value > s_max)
	item_value = s_max;
  else item_value = value;

  page_size = (int) ceil((float)abs(s_max - s_min + 1) / 10.0);

  OnPaint();
  Refresh();
}

void wxSlider::GetSize(int *w, int *h)
{ wxRect srect;

  if(!value_handle || !handle)
	return;

  if(static_handle)
	srect = static_handle->GetWinRect();
  srect = Union(srect, value_handle->GetWinRect());
  srect = Union(srect, handle->GetWinRect());

  *w = srect.W;
  *h = srect.H;
}

void wxSlider::GetPosition(int *x, int *y)
{ wxRect srect;

  if(!value_handle || !handle)
	return;

  if(static_handle)
	srect = static_handle->GetWinRect();
  else  srect = value_handle->GetWinRect();

  ScreenToClient(srect);

  *x = srect.X;
  *y = srect.Y;
}

void wxSlider::OnSetFocus(void)
{
  if(static_handle)
	static_handle->WinAttribOn(A_REVERSE);
  if(value_handle)
	value_handle->WinAttribOn(A_REVERSE);

  OnPaint();
}

void wxSlider::OnKillFocus(void)
{
  if(static_handle)
	static_handle->WinAttribOff(A_REVERSE);
  if(value_handle)
	value_handle->WinAttribOff(A_REVERSE);

  OnPaint();
}

void wxSlider::OnPaint(void)
{ wxRect srect;

  if(static_handle && item_label)
  { static_handle->Clear();
    static_handle->Print(0,0, item_label);
  }

  if(value_handle)
  { value_handle->Clear();
    srect = value_handle->GetWinRect();
    char *tmp = IntToString(item_value);
    value_handle->Print(0,0, tmp);
    delete tmp;
  }

  if(handle)
  { handle->Clear();
    srect = handle->GetWinRect();
    int s_pos = (int) floor((float)item_value / (float)abs(s_max-s_min+1) * (float) (srect.W-2));
    handle->Print(0,0, '<');
    if(s_pos > 0)
	handle->Printf(A_REVERSE, 1,0, "%-*s", s_pos, " ");
    handle->Print(srect.W-1,0,'>');
  }
}

Bool wxSlider::ProcessEvent(wxEvent *ev)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;
  Bool handled = TRUE;

  if (k_ev->eventType == wxEVENT_TYPE_CHAR)
  { switch(k_ev->keyCode)
    {
      case KEY_DOWN:    SetValue(item_value - 1);
			Refresh();
			break;
      case KEY_NPAGE:	SetValue(item_value - page_size);
			Refresh();
			break;
      case KEY_UP:	SetValue(item_value + 1);
			Refresh();
			break;
      case KEY_PPAGE:	SetValue(item_value + page_size);
			Refresh();
			break;
      case KEY_HOME:	SetValue(s_min);
			Refresh();
			break;
      case KEY_END:	SetValue(s_max);
			Refresh();
			break;
      default:		handled = FALSE;
    }

    if(handled)
    { wxCommandEvent event(wxEVENT_TYPE_SLIDER_COMMAND);
      event.commandInt = GetValue();
      event.eventObject = this;
      ProcessCommand(event);
    }
    return handled;
  };
  
  return wxItem::ProcessEvent(ev);
}
