/*
 * File:     wx_panel.cc
 * Purpose:  wxPanel class implementation
 *
 *                       wxWindows 1.40
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 18-4-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "common.h"
#include "wx_panel.h"
#include "wx_frame.h"
#include "wx_utils.h"
#include "wx_buttn.h"

// Constructor

wxPanel::wxPanel(void)
{
  frozen = FALSE;
  cursor_x = 0;
  cursor_y = 0;
  max_height = 0;
  max_line_height = 0;
  max_width = 0;
  hspacing = 1;
  vspacing = 0;
  new_line = FALSE;
}

wxPanel::wxPanel(wxFrame *frame, int x, int y, int width, int height, int style, Bool create):
  wxbPanel(frame, x, y, width, height, style)
{
  frozen = FALSE;
  cursor_x = 0;
  cursor_y = 0;
  max_height = 0;
  max_line_height = 0;
  max_width = 0;
  hspacing = 1;
  vspacing = 0;
  new_line = FALSE;

  if(create)
	Create(frame, x, y, width, height, style);
}

Bool wxPanel::Create(wxFrame *frame, int x, int y, int width, int height, int style)
{ int pw, ph;

  if (!frame) return FALSE;
  window_parent = frame;

  if(x == -1)
	x = 0;
  if(y == -1)
	y = 0;

  frame->GetClientSize(&pw, &ph);
  if (width == -1)
	width = pw;
  if (height == -1)
	height = ph;

  wxRect wrect(x, y, width ,height);
  ClientToScreen(wrect);

  if (style & wxBORDER)
  {
    handle = new CursWnd(frame->handle, wrect, CW_THIN_LINE, this);
    handle->Border(CW_THIN_LINE);
  }
  else handle = new CursWnd(frame->handle, wrect, CW_NO_LINE, this);

  if(handle)
  { frame->AddChild(this);
    handle->Show(TRUE);
    return TRUE;
  }
  else return FALSE;
}

wxPanel::~wxPanel(void)
{
}

void wxPanel::OnSetFocus(void)
{ wxWindow *win;

  win = FocusedSubWin();
  if(win)
    win->SetFocus();
}

Bool wxPanel::ProcessEvent(wxEvent *ev)
{ wxKeyEvent *k_ev = (wxKeyEvent *) ev;

  if (k_ev->eventType == wxEVENT_TYPE_CHAR)
  { switch(k_ev->keyCode)
    { case KEY_RIGHT:	// fall through
      case KEY_DOWN:
      case KEY_TAB:	FocusNextSubWin();
			return TRUE;
      case KEY_UP:
      case KEY_BTAB:
      case KEY_LEFT:	FocusPreviousSubWin();
			return TRUE;
/* Default buttons appear to be unuseable with a character interface
      case KEY_ENTER:	if(defaultItem)
				OnDefaultAction(defaultItem);
			return TRUE;
*/
      default:		return FALSE;
    }
  };

  return wxWindow::ProcessEvent(ev);
}

/*****************************************************************
 * ITEM PLACEMENT FUNCTIONS
 *****************************************************************/


// Start a new line
void wxPanel::NewLine(void)
{
  cursor_x = 0;
  if (max_line_height == 0)
  {
    cursor_y += vspacing;
  }
  else
    cursor_y += vspacing + max_line_height;
  max_line_height = 0;
  if(cursor_y > max_height)
	max_height = cursor_y;
}

void wxPanel::Tab(int pixels)
{
  cursor_x += pixels;
  if (cursor_x > max_width)
    max_width = cursor_x;
}

void wxPanel::GetCursor(int *x, int *y)
{
  *x = cursor_x;
  *y = cursor_y;
}

void wxPanel::SetCursor(int x, int y)
{
  cursor_x = x;
  cursor_y = y;

  if(cursor_x > max_width)
	max_width = cursor_x;
  if(cursor_y > max_height)
	max_height = cursor_y;
}

void wxPanel::Freeze(Bool flag)
{
  frozen = flag;
}

// Set/get horizontal spacing
void wxPanel::SetHorizontalSpacing(int sp)
{
  hspacing = sp;
}

int wxPanel::GetHorizontalSpacing(void)
{
  return hspacing;
}

// Set/get vertical spacing
void wxPanel::SetVerticalSpacing(int sp)
{
  vspacing = sp;
}

int wxPanel::GetVerticalSpacing(void)
{
  return vspacing;
}

// Fits the panel around the items
void wxPanel::Fit(void)
{
  if((max_width != 0) && (max_height != 0))
  	SetClientSize(max_width, max_height);
}

// Update next cursor position
void wxPanel::AdvanceCursor(wxWindow *item)
{
  if(frozen)
	return;

  int width, height;
  int x, y;
  item->GetSize(&width, &height);
  item->GetPosition(&x, &y);

  if ((x + width) > max_width)
    max_width = x + width;
  if ((y + height) > max_height)
    max_height = y + height;
  if (height > max_line_height)
    max_line_height = height;

  cursor_x = x + width + hspacing;
  cursor_y = y;
}

// If x or y are not specified (i.e. < 0), supply
// values based on left to right, top to bottom layout.
// Internal use only.
void wxPanel::GetValidPosition(int *x, int *y)
{
  if (*x < 0)
    *x = cursor_x;

  if (*y < 0)
    *y = cursor_y;
}

