/*
 * File:     wx_win.cc
 * Purpose:  wxWindow class implementation
 *
 *                       wxWindows 1.40
 * Copyright (c) 1993 Artificial Intelligence Applications Institute,
 *                   The University of Edinburgh
 *
 *                     Author: Julian Smart
 *                       Date: 18-4-93
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice, author statement and this permission
 * notice appear in all copies of this software and related documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS,
 * IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL THE ARTIFICIAL INTELLIGENCE APPLICATIONS INSTITUTE OR THE
 * UNIVERSITY OF EDINBURGH BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF
 * DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "common.h"
#include "wx_win.h"
#include "winmgr.h"
#include "wx_main.h"
#include "wx_utils.h"
#include "wx_rect.h"

wxWindow *wxFocusWnd = NULL;
wxFrame *wxFocusFrame = NULL;

// INTERNAL MEMBER FUNCTIONS
void wxWindow::AddDisposal(wxObject *disposable)
{
  if(!disposal)
	disposal = new wxList();
  if(disposal) disposal->Append(disposable);
}

void wxWindow::RemoveDisposal(void)
{
  if(disposal)
  { disposal->DeleteContents(TRUE);
    delete disposal;
    disposal = NULL;
  }
}

wxRect wxWindow::GetWinRect(void)
{
  return handle->GetWinRect();
}

wxRect wxWindow::MinimalWinSize(void)
{ // default minimal size 1x1
  wxRect minrect;

  if(handle->BorderStyle)
	minrect.W = minrect.H = 3;
  else minrect.W = minrect.H = 1;
  return minrect;
}

wxRect wxWindow::GetClientRect(void)
{
  return handle->GetClientRect();
}

void wxWindow::ValidateWinRect(wxRect& rect)
{
   wxRect minrect(MinimalWinSize());
   if(rect.W < minrect.W)
	rect.W = minrect.W;
   if(rect.H < minrect.H)
	rect.H = minrect.H;
}

void wxWindow::ScreenToClient(wxRect &winRect)
{wxWindow *parent = GetParent();
 wxRect prect;

 if (parent)
  {
	prect = parent->GetClientRect();
	winRect.X = winRect.X - prect.X;
	winRect.Y = winRect.Y - prect.Y;
  }
}

void wxWindow::ClientToScreen(wxRect &winRect)
{ wxWindow *parent = GetParent();
  wxRect prect;
  // Since we now have the relative screen coords,
  // if there's a parent we must add its top left corner parent
  if (parent)
  { prect = parent->GetClientRect();
	winRect.X = prect.X + winRect.X;
	winRect.Y = prect.Y + winRect.Y;
  };
}

void wxWindow::Refresh(void)
{
  TheScreen->Refresh();
}

Bool wxWindow::PreProcessEvent(wxEvent *ev)
{
  return FALSE;
}

Bool wxWindow::ProcessEvent(wxEvent *ev)
{ 
  if(ev->eventType == wxEVENT_TYPE_WM_COMMAND)
  { wxCommandEvent *c_ev = (wxCommandEvent *) ev;
    switch(c_ev->commandInt)
    { 
      case WM_HELP:             return OnHelp();
    }
  }

  return FALSE;
}

Bool wxWindow::PostProcessEvent(wxEvent *ev, Bool handled)
{
  return handled;
}

wxWindow* wxWindow::FocusedSubWin(void)
{  wxNode *node;
   CursWnd *cwin;
   wxWindow *win;

   if(!handle) return NULL;

   node = handle->Children.Last();

   while(node)
   { cwin = (CursWnd *) node->Data();
     win = cwin->wx_window;
     if(win && win->WantFocus()) return win;
     node = node->Previous();
   }
   return NULL;
}

Bool wxWindow::InFocusChain(void)
{ wxNode *node;
  wxWindow *win;

  if((wxFocusWnd == this) || (wxFocusFrame == (wxFrame *) this))
	return TRUE;
  for(node = children->First();node;node = node->Next())
  { win = (wxWindow *) node->Data();
    if(win->InFocusChain())
	return TRUE;
  }
  return FALSE;
}

Bool wxWindow::WantFocus(void)
{
  return wantsFocus;
}

void wxWindow::WantFocus(Bool flag)
{
  wantsFocus = flag;
}

Bool wxWindow::Visible(void)
{
  if(!handle)
	return FALSE;
  else return handle->Visible;
}

void wxWindow::FocusPreviousFrame(void)
{ wxNode *node, *thisnode;
  wxFrame *next_swin;
  Bool found = FALSE;

  // Cycle through the windows
  thisnode = wxTopLevelWindows.Member((wxObject *) this);
  if(!thisnode)
  { // The current focus frame is not valid any more
    wxFocusFrame = NULL;
    // The current focus wnd is not valid any more
    wxFocusWnd = NULL;
    return;
  }

  // Get the previous item
  node = thisnode->Previous();
  if(!node) node = wxTopLevelWindows.Last();

  while(node)
  {	
	if(thisnode == node) return;

	next_swin = (wxFrame *) node->Data();
	if(next_swin->Visible() && next_swin->WantFocus())
	{ next_swin->SetFocus();
          found = TRUE;
	  break;
	}
	node = node->Previous();
	if(!node) node = wxTopLevelWindows.Last();
  }

  if(!found)
  { // The current focus frame is not valid any more
    wxFocusFrame = NULL;
    // The current focus wnd is not valid any more
    wxFocusWnd = NULL;
  }

  Refresh();
}

void wxWindow::FocusPreviousWin(void)
{ wxWindow *parent = GetParent();

  if(parent)
	parent->FocusPreviousSubWin();
  else FocusPreviousFrame();
}

void wxWindow::FocusNextSubWin(void)
{ wxNode *thisnode, *node;
  wxWindow *next_swin;
  wxWindow *focus_swin;
  Bool found = FALSE;

  focus_swin = FocusedSubWin();
  if (!focus_swin)
  { // The current focus wnd is not valid any more
    wxFocusWnd = NULL;
    return;
  }

  // Cycle through the windows
  thisnode = children->Member((wxObject *) focus_swin);
  if (!thisnode)
  { // The current focus wnd is not valid any more
    wxFocusWnd = NULL;
    return;
  }

  node = thisnode->Next();   // Get the next item
  if(!node) node = children->First();

  while(node)
  {
	if(thisnode == node) return;

	next_swin = (wxWindow *) node->Data();
	if(next_swin->Visible() && next_swin->WantFocus())
	{ next_swin->SetFocus();
	  found = TRUE;
	  break;
	}
	else
	node = node->Next();
	if(!node) node = children->First();
  }

  if(!found)
  { // The current focus wnd is not valid any more
    wxFocusWnd = NULL;
  }

  Refresh();
}

void wxWindow::FocusPreviousSubWin(void)
{ wxNode *node, *thisnode;
  wxWindow *next_swin;
  wxWindow *focus_swin;
  Bool found = FALSE;

  focus_swin = FocusedSubWin();
  if (!focus_swin)
  { // The current focus wnd is not valid any more
    wxFocusWnd = NULL;
    return;
  }

// Cycle through the windows
  thisnode = children->Member((wxObject *) focus_swin);
  if(!thisnode)
  { // The current focus wnd is not valid any more
    wxFocusWnd = NULL;
    return;
  }

  node = thisnode->Previous();
  if(!node) node = children->Last();  // was it the first ?

  while(node)
  {	
	if(thisnode == node) return;

	next_swin = (wxWindow *) node->Data();
	if(next_swin->Visible() && next_swin->WantFocus())
	{ next_swin->SetFocus();
	  found = TRUE;
	  break;
	}
	node = node->Previous();
	if(!node) node = children->Last();
  }

  if(!found)
  { // The current focus wnd is not valid any more
    wxFocusWnd = NULL;
  }

  Refresh();
}

void wxWindow::KillFocus(void)
{
  if((InFocusChain()))
	FocusPreviousWin();

  if(this == wxFocusWnd)
	wxFocusWnd = NULL;
  if(((wxFrame *) this) == wxFocusFrame)
  { wxFocusFrame = NULL;
    wxFocusWnd = NULL;
  }
}

void wxWindow::WantPreProcessEvent(wxWindow *win)
{
  if(!preEventHandlers)
	preEventHandlers = new wxList();
  preEventHandlers->Append(win);
}

void wxWindow::WantPostProcessEvent(wxWindow *win)
{
  if(!postEventHandlers)
	postEventHandlers = new wxList();
  postEventHandlers->Append(win);
}

// Constructor
wxWindow::wxWindow(void)
{
  handle = NULL;
  disposal = NULL; // Cleaned up by destructor;contains non-windows etc.
  wantsFocus = TRUE;
  preEventHandlers = NULL;
  postEventHandlers = NULL;
}

// Destructor
wxWindow::~wxWindow(void)
{ wxWindow *parent = GetParent();

  if (parent)
	parent->RemoveChild(this);
  DestroyChildren();
  RemoveDisposal();
  if (handle)
  { delete handle;
    handle = NULL;
  };
  if(preEventHandlers)
  { delete preEventHandlers;
    preEventHandlers = NULL;
  }
  if(postEventHandlers)
  { delete postEventHandlers;
    postEventHandlers = NULL;
  }
  delete children;
  children = NULL;

  Refresh();
}

char *wxWindow::GetHandle(void)
{
  return (char *) handle;
}

// Get total size
void wxWindow::GetSize(int *width, int *height)
{ wxRect rect;

  if(!handle) return;

  rect = handle->GetWinRect();
  *width = rect.W;
  *height = rect.H;
}

void wxWindow::GetPosition(int *x, int *y)
{
  wxRect rect;

  if(!handle) return;

  rect = handle->GetWinRect();
  ScreenToClient(rect);
  *x = rect.X;
  *y = rect.Y;
}

wxCursor *wxWindow::SetCursor(wxCursor *cursor)
{
  wxCursor *old_cursor = wx_cursor;
  wx_cursor = cursor;

/* There are three cursor types:
  1. no cursor
  2. underline cursor
  3. block (insert) cursor
*/

  wxFlushEvents();
  return old_cursor;
}

void wxWindow::SetColourMap(wxColourMap *cmap)
{
}

// Get size *available for subwindows* i.e. excluding menu bar etc.
void wxWindow::GetClientSize(int *width, int *height)
{
  if(!handle) return;

  wxRect crect = GetClientRect();
  *width = crect.W;
  *height = crect.H;
}

void wxWindow::SetSize(int x, int y, int width, int height)
{
  int currentX, currentY;
  int currentW, currentH;
  wxRect winRect(x, y, width, height);

  if(!handle) return;

  GetPosition(&currentX, &currentY);
  if (x == -1)
	winRect.X = currentX;
  if (y == -1)
	winRect.Y = currentY;

  GetSize(&currentW, &currentH);
  if (width  == -1)
	winRect.W = currentW;
  if (height == -1)
	winRect.H = currentH;
 
  ValidateWinRect(winRect);

  // Since we now have the relative screen coords,
  // if there's a parent we must add its top left corner
  ClientToScreen(winRect);

  handle->SetWinRect(winRect);
  Refresh();

  OnSize(width, height);
}

void wxWindow::SetClientSize(int width, int height)
{
  int cur_width, cur_height;
  int cl_width, cl_height;
  int new_width, new_height;

  // Find the difference between the entire window (title bar and all)
  // and the client area; add this to the new client size to move the
  // window

  GetSize(&cur_width, &cur_height);
  GetClientSize(&cl_width, &cl_height);
  new_width = cur_width - cl_width + width;
  new_height = cur_height - cl_height + height;

  SetSize(-1, -1, new_width, new_height);
}

void wxWindow::Show(Bool show)
{
  if(!handle) return;

  handle->Show(show);
  if (show)
  { if(WantFocus())
       SetFocus();
  }
  else KillFocus();

  Refresh();
}

void wxWindow::SetFocus(void)
{
  if(!handle) return;

  if(window_parent)
  { window_parent->handle->SetFocus(handle);
    if(wxFocusWnd)
	wxFocusWnd->OnKillFocus();
    wxFocusWnd = this;
  }
  else
  { TheScreen->SetFocus(handle);
    if(wxFocusFrame)
	wxFocusFrame->OnKillFocus();
    wxFocusFrame = (wxFrame *) this;
    wxTopLevelWindows.DeleteObject(this); // make it the last in the list
    wxTopLevelWindows.Append(this);
  }

  OnSetFocus();
  Refresh();
}

int wxWindow::GetTextFamily(void)
{
// There is only one font family
  return wxDEFAULT;
}

int wxWindow::GetTextStyle(void)
{ int wx_style, win_attrib;

  win_attrib = handle->WinAttrib();
  switch(win_attrib)
  { case A_UNDERLINE: wx_style = wxITALIC;
                      break;
    default         : wx_style = wxDEFAULT;
                      break;
  }

  return wx_style;
}

int wxWindow::GetTextWeight(void)
{ int wx_weight, win_attrib;

  win_attrib = handle->WinAttrib();
  switch(win_attrib)
  { case A_DIM:      wx_weight = wxLIGHT;
		     break;
    case A_BOLD:     wx_weight = wxBOLD;
		     break;
    case A_STANDOUT:
    default:         wx_weight = wxDEFAULT;
		     break;
  }

  return wx_weight;
}

// Assumption. A char has a height and width of 1 pixel
float wxWindow::GetTextHeight(void)
{
  return 1.0;
}

float wxWindow::GetTextWidth(void)
{

  return 1.0;
}

void wxWindow::GetTextExtent(char *string, float *x, float *y)
{
  *x = (float) strlen(string);
  *y = 1.0;
}

// Return the number of chars that this pixel dimension represents
int wxWindow::XChar(int pixels)
{
  return pixels;
}

int wxWindow::YChar(int pixels)
{
  return pixels;
}

Bool wxWindow::OnHelp(void)
{
  return FALSE;
}

void wxWindow::Enable(Bool enable) {}
void wxWindow::CaptureMouse(void) {}
void wxWindow::ReleaseMouse(void) {}
void wxWindow::DragAcceptFiles(Bool accept) {}

